import base64
import hashlib
import json
import random
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

GLOBAL_CONFIG_DIR = Path.home() / ".cligram"
_config_instance: Optional["Config"] = None


class ScanMode(Enum):
    """Operation modes for the scanner."""

    FULL = "full"
    """Full operation mode: scans and sends messages to targets."""

    SCAN = "scan"
    """Scan mode: only scans and stores eligible usernames without sending."""

    SEND = "send"
    """Send mode: only sends messages to eligible usernames without scanning."""

    HALT = "halt"
    """Halt mode: logs in to telegram and shuts down."""

    RECEIVE = "receive"
    """Receive mode: receives and shows new messages."""

    LOGOUT = "logout"
    """Logout mode: logs out from the Telegram and deletes the session file."""


@dataclass
class ApiConfig:
    """Telegram API credentials."""

    id: int = 0
    """Telegram API ID obtained from my.telegram.org/apps"""

    hash: str = ""
    """Telegram API hash string obtained from my.telegram.org/apps"""

    @property
    def identifier(self) -> str:
        """Get unique identifier for the API credentials."""
        hasher = hashlib.sha256()
        hasher.update(f"{self.id}:{self.hash}".encode("utf-8"))
        digest = hasher.digest()
        return base64.urlsafe_b64encode(digest).decode("utf-8")[:8]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ApiConfig":
        return cls(
            id=data.get("id", cls.__dataclass_fields__["id"].default),
            hash=data.get("hash", cls.__dataclass_fields__["hash"].default),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {"id": self.id, "hash": self.hash}


@dataclass
class DelayConfig:
    """Delay interval configuration."""

    min: float = 10.0
    """Minimum delay in seconds"""

    max: float = 20.0
    """Maximum delay in seconds"""

    def random(self) -> float:
        """Generate random delay within configured bounds"""
        return random.uniform(self.min, self.max)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "DelayConfig":
        return cls(
            min=data.get("min", cls.__dataclass_fields__["min"].default),
            max=data.get("max", cls.__dataclass_fields__["max"].default),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {"min": self.min, "max": self.max}


@dataclass
class LongDelayConfig(DelayConfig):
    """Configuration for long delay periods."""

    min: float = 30.0
    max: float = 60.0

    chance: float = 0.1
    """Probability (0-1) of taking a long delay instead of normal delay"""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LongDelayConfig":
        return cls(
            min=data.get("min", cls.__dataclass_fields__["min"].default),
            max=data.get("max", cls.__dataclass_fields__["max"].default),
            chance=data.get("chance", cls.__dataclass_fields__["chance"].default),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {"min": self.min, "max": self.max, "chance": self.chance}


@dataclass
class DelaysConfig:
    """Delay timing configuration."""

    normal: DelayConfig = field(default_factory=DelayConfig)
    """Normal delay settings"""

    long: LongDelayConfig = field(default_factory=LongDelayConfig)
    """Long break delay settings"""

    def random(self) -> float:
        """
        Generate a random delay based on configured normal and long delays.

        Returns:
            float: Random delay duration in seconds
        """
        delay: float
        if random.random() < self.long.chance:
            delay = self.long.random()
        else:
            delay = self.normal.random()

        delay = round(delay, 1)
        return delay

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "DelaysConfig":
        return cls(
            normal=DelayConfig.from_dict(data.get("normal", {})),
            long=LongDelayConfig.from_dict(data.get("long", {})),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {"normal": self.normal.to_dict(), "long": self.long.to_dict()}


@dataclass
class MessagesConfig:
    """Configuration for message forwarding."""

    source: str = "me"
    """Source of messages to forward ('me' or channel username)"""

    limit: int = 20
    """Maximum number of messages to be loaded from source"""

    msg_id: Optional[int] = None
    """Specific message ID to forward (optional)"""

    @property
    def randomize(self) -> bool:
        """Determine if message selection should be randomized."""
        return self.msg_id is None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "MessagesConfig":
        return cls(
            source=data.get("source", cls.__dataclass_fields__["source"].default),
            limit=data.get("limit", cls.__dataclass_fields__["limit"].default),
            msg_id=data.get("msg_id", cls.__dataclass_fields__["msg_id"].default),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {"source": self.source, "limit": self.limit, "msg_id": self.msg_id}


@dataclass
class ScanConfig:
    """Configuration for scanning behavior and timing."""

    messages: MessagesConfig = field(default_factory=MessagesConfig)
    """Message forwarding settings"""

    mode: ScanMode = ScanMode.FULL
    """Operation mode"""

    targets: List[str] = field(default_factory=list)  # type: ignore
    """List of target groups to scan (usernames or URLs)"""

    limit: int = 50
    """Maximum number of messages to scan per group"""

    test: bool = False
    """Test mode without sending messages"""

    rapid_save: bool = False
    """Enable rapid state saving to disk"""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ScanConfig":
        return cls(
            messages=MessagesConfig.from_dict(data.get("messages", {})),
            mode=ScanMode(
                data.get("mode", cls.__dataclass_fields__["mode"].default.value)
            ),
            targets=data.get(
                "targets", cls.__dataclass_fields__["targets"].default_factory()  # type: ignore
            ),
            limit=data.get("limit", cls.__dataclass_fields__["limit"].default),
            test=data.get("test", cls.__dataclass_fields__["test"].default),
            rapid_save=data.get(
                "rapid_save", cls.__dataclass_fields__["rapid_save"].default
            ),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "messages": self.messages.to_dict(),
            "mode": self.mode.value,
            "targets": self.targets,
            "limit": self.limit,
            "test": self.test,
            "rapid_save": self.rapid_save,
        }


@dataclass
class ConnectionConfig:
    """Connection settings for Telegram client."""

    direct: bool = True
    """Whether to allow direct connection"""

    proxies: List[str] = field(default_factory=list)  # type: ignore
    """List of proxy URLs to try for connection"""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ConnectionConfig":
        return cls(
            direct=data.get(
                "direct",
                cls.__dataclass_fields__["direct"].default,
            ),
            proxies=data.get(
                "proxies", cls.__dataclass_fields__["proxies"].default_factory()  # type: ignore
            ),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "direct": self.direct,
            "proxies": self.proxies,
        }


@dataclass
class StartupConfig:
    """Telegram client startup settings."""

    count_unread_messages: bool = True
    """Show unread messages count on startup"""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "StartupConfig":
        return cls(
            count_unread_messages=data.get(
                "count_unread_messages",
                cls.__dataclass_fields__["count_unread_messages"].default,
            ),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "count_unread_messages": self.count_unread_messages,
        }


@dataclass
class TelegramConfig:
    """Telegram client settings."""

    api: ApiConfig = field(default_factory=ApiConfig)
    """API credentials from my.telegram.org"""

    connection: ConnectionConfig = field(default_factory=ConnectionConfig)
    """Connection settings"""

    startup: StartupConfig = field(default_factory=StartupConfig)
    """Startup behavior settings"""

    session: str = "default"
    """Session file name for persistent authorization"""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "TelegramConfig":
        return cls(
            api=ApiConfig.from_dict(data.get("api", {})),
            connection=ConnectionConfig.from_dict(data.get("connection", {})),
            startup=StartupConfig.from_dict(data.get("startup", {})),
            session=data.get("session", cls.__dataclass_fields__["session"].default),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "api": self.api.to_dict(),
            "connection": self.connection.to_dict(),
            "startup": self.startup.to_dict(),
            "session": self.session,
        }


@dataclass
class AppConfig:
    """Main application behavior configuration."""

    delays: DelaysConfig = field(default_factory=DelaysConfig)
    """Delay timing configurations"""

    verbose: bool = False
    """Enable debug logging"""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "AppConfig":
        return cls(
            delays=DelaysConfig.from_dict(data.get("delays", {})),
            verbose=data.get("verbose", cls.__dataclass_fields__["verbose"].default),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "delays": self.delays.to_dict(),
            "verbose": self.verbose,
        }


class InteractiveMode(Enum):
    """Interactive mode options."""

    CLIGRAM = "cligram"
    """Interactive mode with Cligram commands"""

    PYTHON = "python"
    """Interactive mode with Python code execution"""


@dataclass
class InteractiveConfig:
    """Interactive mode configuration."""

    mode: InteractiveMode = InteractiveMode.CLIGRAM
    """The interactive mode to use"""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "InteractiveConfig":
        return cls(
            mode=InteractiveMode(
                data.get("mode", cls.__dataclass_fields__["mode"].default.value)
            )
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "mode": self.mode.value,
        }


@dataclass
class Config:
    """Application configuration root."""

    app: AppConfig = field(default_factory=AppConfig)
    """Application behavior settings"""

    telegram: TelegramConfig = field(default_factory=TelegramConfig)
    """Telegram client and connection settings"""

    scan: ScanConfig = field(default_factory=ScanConfig)
    """Scanning behavior and timing settings"""

    interactive: InteractiveConfig = field(default_factory=InteractiveConfig)
    """Interactive mode settings"""

    exclusions: List[str] = field(default_factory=list)
    """List of usernames to exclude from processing"""

    path: Path = field(default=Path("config.json"))
    """Path to the configuration file"""

    updated: bool = False
    """Indicates if the configuration was updated with new fields"""

    @property
    def base_path(self) -> Path:
        """Get base directory of the configuration file."""
        return self.path.parent

    @property
    def data_path(self) -> Path:
        """Get data directory path."""
        return self.base_path / "data"

    @classmethod
    def get_config(cls, raise_if_failed: bool = True) -> "Config":
        """Get application configurations."""
        if raise_if_failed:
            if _config_instance is None:
                raise RuntimeError("Configuration not loaded. Call from_file() first.")
            if not isinstance(_config_instance, cls):
                raise TypeError("Configuration instance is of incorrect type.")

        return _config_instance if isinstance(_config_instance, cls) else None  # type: ignore

    @classmethod
    def from_file(
        cls,
        config_path: str | Path = "config.json",
        overrides: Optional[List[str]] = None,
    ) -> "Config":
        """Load configuration from JSON file with CLI overrides."""
        config_full_path = Path(config_path).resolve()
        if not config_full_path.exists():
            raise FileNotFoundError(f"Configuration file not found: {config_path}")

        with open(config_full_path, "r") as f:
            original_data = json.load(f)

        # Parse main sections
        config = cls(
            app=AppConfig.from_dict(original_data.get("app", {})),
            telegram=TelegramConfig.from_dict(original_data.get("telegram", {})),
            scan=ScanConfig.from_dict(original_data.get("scan", {})),
            interactive=InteractiveConfig.from_dict(
                original_data.get("interactive", {})
            ),
            path=config_full_path,
        )

        # Apply overrides
        if overrides:
            for override in overrides:
                config.apply_override(override)

        # Check if config structure changed (new fields added)
        new_data = config.to_dict()
        if not cls._config_equal(original_data, new_data):
            config._update_config(original_data)
            config.updated = True

        if not cls.get_config(raise_if_failed=False):
            global _config_instance
            _config_instance = config

        return config

    def apply_override(self, override_str: str):
        """
        Apply a configuration override using dot notation.

        Args:
            override_str: Override string in format "path.to.key=value"
                         Examples: "app.verbose=true", "scan.limit=200"

        Raises:
            ValueError: If override string is invalid
        """
        if "=" not in override_str:
            raise ValueError(
                f"Invalid override format: {override_str}. Expected 'key=value'"
            )

        path, value_str = override_str.split("=", 1)
        path = path.strip()

        # Parse value
        value = self._parse_value(value_str)

        # Apply override
        self._set_nested_value(path, value)

    def to_dict(self) -> Dict[str, Any]:
        """Export configuration as dictionary."""
        return {
            "app": self.app.to_dict(),
            "telegram": self.telegram.to_dict(),
            "scan": self.scan.to_dict(),
            "interactive": self.interactive.to_dict(),
        }

    def save(self, path: Optional[Path | str] = None):
        """Save configuration to JSON file."""
        save_path = Path(path) if path else self.path
        save_path.parent.mkdir(parents=True, exist_ok=True)

        with open(save_path, "w") as f:
            json.dump(self.to_dict(), f, indent=2)

    def _parse_value(self, value_str: str) -> Any:
        """Parse string value to appropriate Python type."""
        # Boolean
        value_str = value_str.strip()

        if value_str.lower() in ("true", "yes", "1"):
            return True

        if value_str.lower() in ("false", "no", "0"):
            return False

        # None/null
        if value_str.lower() in ("none", "null"):
            return None

        # Number
        try:
            if "." in value_str:
                return float(value_str)
            return int(value_str)
        except ValueError:
            pass

        # List (JSON array)
        if value_str.startswith("[") and value_str.endswith("]"):
            try:
                return json.loads(value_str)
            except json.JSONDecodeError:
                pass

        # String (remove quotes if present)
        if (value_str.startswith('"') and value_str.endswith('"')) or (
            value_str.startswith("'") and value_str.endswith("'")
        ):
            return value_str[1:-1]

        return value_str

    def _set_nested_value(self, path: str, value: Any):
        """
        Set a nested configuration value using dot notation.

        Args:
            path: Dot-separated path to value (e.g., "app.verbose")
            value: Value to set

        Raises:
            ValueError: If path is invalid
        """
        parts = path.split(".")

        if len(parts) < 2:
            raise ValueError(f"Invalid path: {path}. Must have at least one dot.")

        # Navigate to parent object
        obj = self
        for part in parts[:-1]:
            if not hasattr(obj, part):
                raise ValueError(f"Invalid path: {path}. '{part}' not found.")
            obj = getattr(obj, part)

        # Set the final value
        attr = parts[-1]
        if not hasattr(obj, attr):
            raise ValueError(f"Invalid path: {path}. '{attr}' not found.")

        # Type conversion for enums
        if hasattr(obj.__class__, "__dataclass_fields__"):
            field_info = obj.__class__.__dataclass_fields__.get(attr)
            if field_info and field_info.type == ScanMode:
                value = ScanMode(value)

        setattr(obj, attr, value)

    def get_nested_value(self, path: str) -> Any:
        """
        Get a nested configuration value using dot notation.

        Args:
            path: Dot-separated path to value (e.g., "app.verbose")

        Returns:
            Value at the specified path

        Raises:
            ValueError: If path is invalid
        """
        parts = path.split(".")
        obj = self

        for part in parts:
            if not hasattr(obj, part):
                raise ValueError(f"Invalid path: {path}. '{part}' not found.")
            obj = getattr(obj, part)

        return obj

    @staticmethod
    def _flatten_dict(
        d: Dict[str, Any], parent_key: str = "", sep: str = "."
    ) -> Dict[str, Any]:
        """
        Flatten a nested dictionary.

        Args:
            d: Dictionary to flatten
            parent_key: Parent key prefix
            sep: Separator for nested keys

        Returns:
            Flattened dictionary with dot notation keys
        """
        items: List[Tuple[str, Any]] = []
        for k, v in d.items():
            new_key = f"{parent_key}{sep}{k}" if parent_key else k
            if isinstance(v, dict):
                items.extend(Config._flatten_dict(v, new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)

    @staticmethod
    def _config_equal(old: Dict[str, Any], new: Dict[str, Any]) -> bool:
        """
        Compare two configuration dictionaries for structural equality.

        Returns True if they have the same structure (keys), ignoring values.
        This detects when new fields are added to the config schema.
        """
        # Flatten both dicts
        flat_old = Config._flatten_dict(old)
        flat_new = Config._flatten_dict(new)

        # Compare keys only
        return set(flat_old.keys()) == set(flat_new.keys())

    def _update_config(self, old_data: Dict[str, Any]):
        # Migrate existing config keys to new structure
        if old_data.get("app", {}).get("rapid_save") is not None:
            self.scan.rapid_save = old_data["app"]["rapid_save"]
        if old_data.get("telegram", {}).get("proxies") is not None:
            self.telegram.connection.proxies = old_data["telegram"]["proxies"]
        if old_data.get("telegram", {}).get("direct_connection") is not None:
            self.telegram.connection.direct = old_data["telegram"]["direct_connection"]

        new_data = self.to_dict()

        # Create backup
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        backup_path = (
            self.path.parent / f"{self.path.stem}.backup.{timestamp}{self.path.suffix}"
        )

        with open(backup_path, "w") as f:
            json.dump(old_data, f, indent=2)

        # Save updated config
        with open(self.path, "w") as f:
            json.dump(new_data, f, indent=2)


def get_search_paths() -> List[Path]:
    """Get a list of all configuration search paths."""
    return [Path.cwd(), GLOBAL_CONFIG_DIR]


def find_config_file(raise_error: bool = False) -> Optional[Path]:
    """Search for configuration file in standard locations."""
    search_paths = get_search_paths()
    config_filenames = ["config.json", "cligram_config.json"]

    for search_dir in search_paths:
        for filename in config_filenames:
            candidate = search_dir / filename
            if candidate.exists():
                return candidate.resolve()

    if raise_error:
        raise FileNotFoundError("No configuration file found in standard locations.")

    return None



import asyncio
import copy
import hashlib
import json
import logging
import os
import shutil
from abc import ABC, abstractmethod
from argparse import ArgumentError
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional, Set, TypeVar

logger: logging.Logger = None  # type: ignore


class State(ABC):
    """
    Abstract base class for persistent state storage.

    Defines interface for loading, saving, and tracking changes
    to state data with support for validation and atomic operations.
    """

    @abstractmethod
    def load(self, path: str) -> None:
        """
        Load state data from file.

        Args:
            path: Path to state file
        """
        pass

    @abstractmethod
    def save(self, path: str) -> None:
        """
        Save state data to file.

        Args:
            path: Path to state file
        """
        pass

    @abstractmethod
    def changed(self) -> bool:
        """Check if state changed since last save."""
        pass


StateT = TypeVar("StateT", bound=State)


class JsonState(State):
    """
    JSON-based state persistence implementation.

    Features:
    - Atomic file operations
    - Schema validation
    - Change tracking
    - Data corruption detection
    - Set/list type conversion
    """

    def __init__(self):
        self._default_data: Dict[str, Any] = getattr(self, "_default_data", {})
        self.data: Dict[str, Any] = copy.deepcopy(self._default_data)
        self.schema: Optional[Dict[str, Any]] = None
        self.corrupted: bool = False
        self._should_save: bool = False
        self._last_hash = self.get_hash()

    def _sets_to_lists(self, data: Any) -> Any:
        """
        Convert set objects to lists for JSON serialization.

        Args:
            data: Data structure containing sets

        Returns:
            Data structure with sets converted to lists
        """
        if isinstance(data, dict):
            return {k: self._sets_to_lists(v) for k, v in data.items()}
        elif isinstance(data, set):
            return list(data)
        elif isinstance(data, list):
            return [self._sets_to_lists(item) for item in data]
        return data

    def load(self, path: str) -> None:
        """Load state data from JSON file."""
        if self.changed():
            raise RuntimeError("Cannot load state with unsaved changes")

        data = None

        try:
            data = self._read_json(path)

            if data is None:
                self._should_save = True
                logger.warning(f"No data found at {path}")
                return

            if not isinstance(data, dict):
                raise ValueError(
                    f"Invalid data format in {path}: expected dict, got {type(data).__name__}"
                )
        except:
            self.corrupted = True
            raise

        # merge with default data
        data = copy.deepcopy(self._default_data) | data

        if self.schema:
            if not self.verify_structure(data, self.schema):
                self.corrupted = True
                raise ValueError(f"Invalid structure in {path}")
        self.data = data
        self._last_hash = self.get_hash()

    def save(self, path: str) -> None:
        """Save state data to JSON file."""
        if self.corrupted:
            raise RuntimeError("Cannot save corrupted state")

        if self.changed():
            if self.schema:
                if not self.verify_structure(self.data, self.schema):
                    path = f"{path}.corrupted"
                    logger.warning(f"Invalid structure, saving to {path}")
            self._atomic_save(path)
            self._should_save = False
            self._last_hash = self.get_hash()

    def changed(self) -> bool:
        """Check if state data has changed since last save."""
        return not self.corrupted and (
            self._should_save
            or self._last_hash is None
            or self.get_hash() != self._last_hash
        )

    def get_hash(self) -> str:
        """Get a hash of the current state data."""
        m = hashlib.sha256()
        json_data = self._sets_to_lists(self.data)
        m.update(json.dumps(json_data, sort_keys=True).encode("utf-8"))
        return m.hexdigest()

    def _read_json(self, path: str) -> Optional[Dict[str, Any]]:
        """Read JSON data from file."""
        try:
            with open(path, "r") as f:
                return json.load(f)
        except FileNotFoundError:
            return None

    def _atomic_save(self, path: str) -> None:
        """Save state data to JSON file atomically."""
        tmp = f"{path}.tmp"
        try:
            # Convert sets to lists before saving
            json_data = self._sets_to_lists(self.data)
            with open(tmp, "w") as f:
                json.dump(json_data, f, indent=2)
            os.replace(tmp, path)
        except Exception:
            if os.path.exists(tmp):
                os.unlink(tmp)
            raise

    @classmethod
    def verify_structure(
        cls, data: Any, schema: Dict[str, Any], path: str = ""
    ) -> bool:
        """
        Recursively verify that data matches the provided schema.
        Returns True if valid, False otherwise.
        Logs warnings for mismatches.
        """
        if isinstance(schema, dict):
            if not isinstance(data, dict):
                logger.warning(
                    f"Structure mismatch at {path or 'root'}: expected dict, got {type(data).__name__}"
                )
                return False
            for key, subschema in schema.items():
                if key not in data:
                    logger.warning(f"Missing key '{key}' at {path or 'root'}")
                    return False
                if not cls.verify_structure(
                    data[key], subschema, path=f"{path}.{key}" if path else key
                ):
                    return False
            return True
        elif isinstance(schema, list):
            if not isinstance(data, list):
                logger.warning(
                    f"Structure mismatch at {path or 'root'}: expected list, got {type(data).__name__}"
                )
                return False
            if schema:
                subschema = schema[0]
                for idx, item in enumerate(data):
                    if not cls.verify_structure(item, subschema, path=f"{path}[{idx}]"):
                        return False
            return True
        elif isinstance(schema, type):
            if not isinstance(data, schema):
                logger.warning(
                    f"Type mismatch at {path or 'root'}: expected {schema.__name__}, got {type(data).__name__}"
                )
                return False
            return True
        else:
            logger.warning(f"Unknown schema type at {path or 'root'}: {schema}")
            return False


class UsersState(JsonState):
    """
    User state tracking implementation.

    Maintains:
    - Set of messaged user IDs
    - Set of eligible usernames
    - Automatic type conversion
    """

    def __init__(self):
        self._default_data = {
            "messaged": set(),
            "eligible": set(),
        }

        super().__init__()

    def load(self, path) -> None:
        """Load user state data from JSON file."""
        super().load(path)

        if not isinstance(self.data["messaged"], set):
            self.data["messaged"] = set(self.data["messaged"])
        if not isinstance(self.data["eligible"], set):
            self.data["eligible"] = set(self.data["eligible"])

        self._last_hash = self.get_hash()

    @property
    def messaged(self) -> Set[int]:
        """Get the set of users that have been messaged."""
        return self.data["messaged"]

    @property
    def eligible(self) -> Set[str]:
        """Get the set of eligible usernames."""
        return self.data["eligible"]


@dataclass
class GroupInfo:
    """
    Represents a chat group's scanning window state.

    The scanning window defines the range of message IDs that have
    been processed, allowing for incremental scanning of large groups.

    On any attribute update (except 'id'), the parent GroupsState's data is updated.
    """

    id: str
    """Group's unique identifier"""

    max: Optional[int] = None
    """Highest message ID scanned"""

    min: Optional[int] = None
    """Lowest message ID scanned"""

    _parent: Optional["GroupsState"] = None
    """Reference to parent GroupsState (not serialized)"""

    def __setattr__(self, name, value):
        if name in ["id", "_parent"]:
            if getattr(self, name, None) is not None:
                raise AttributeError(
                    f"Cannot change '{name}' attribute after initialization"
                )

        super().__setattr__(name, value)
        # Update parent GroupsState's data dict on any change except 'id' and '_parent'
        if (
            name not in ("id", "_parent")
            and hasattr(self, "_parent")
            and self._parent is not None
        ):
            self._parent._update_group_data(self.id, name, value)

    def update(self) -> None:
        """
        Update the group info in the parent GroupsState.

        This method is called to ensure the parent state is updated
        with the current values of this GroupInfo instance.
        """
        if self._parent is not None:
            self._parent.update(self)
        else:
            raise ReferenceError(
                "Cannot update GroupInfo without parent GroupsState reference"
            )


class GroupsState(JsonState):
    """
    Group scanning state implementation.

    Tracks:
    - Message ID windows per group
    - Incremental scanning progress
    - Group metadata

    Keeps GroupInfo and internal data dict in sync on attribute changes.
    """

    def __init__(self):
        super().__init__()

        self._groups: Dict[str, GroupInfo] = {}

    def _update_group_data(self, group_id: str, attr: str, value: Any):
        """Update the internal data dict for a group."""
        if group_id not in self.data:
            self.data[group_id] = {}
        self.data[group_id][attr] = value

    def get(self, group_id: str) -> GroupInfo:
        """Get a group by ID, ensuring live sync with internal data."""
        if group_id in self._groups:
            return self._groups[group_id]

        data = {}
        if group_id in self.data:
            data = self.data[group_id]

        group = GroupInfo(id=group_id, _parent=self, **data)
        self._groups[group_id] = group
        return group

    def update(self, group: GroupInfo) -> None:
        """
        Update or add a group.

        Args:
            group: GroupInfo instance with updated data
        """
        if not isinstance(group, GroupInfo):
            raise TypeError("Expected GroupInfo instance")

        self._groups[group.id] = group
        for key, value in group.__dict__.items():
            if key not in ["id", "_parent"]:
                self._update_group_data(group.id, key, value)


class StateManager:
    """
    Manages persistent application state and handles file-based storage operations.

    Responsible for:
    - Tracking messaged and eligible users
    - Managing group scanning windows
    - Saving/loading state from disk
    - Creating backups of state files
    """

    def __init__(self, data_dir: str | Path, backup_dir: Optional[str | Path] = None):
        """
        Initialize the state manager.

        Args:
            data_dir (str): Path to the data directory storing all state files
        """
        global logger
        logger = logging.getLogger("cligram.state_manager")

        self.data_dir = Path(data_dir).resolve()
        self.data_dir.mkdir(parents=True, exist_ok=True)

        self.backup_dir = (
            Path(backup_dir).resolve() if backup_dir else self.data_dir / "backup"
        )
        self._need_backup = False

        self.lock = asyncio.Lock()

        self.states: Dict[str, State] = {}

        # Initialize core states
        self.users: UsersState = self.register("users", UsersState())
        self.groups: GroupsState = self.register("groups", GroupsState())

    def _get_state_path(self, name: str) -> Path:
        """Get full path for state file."""
        return self.data_dir / f"{name}.json"

    def register(self, name: str, state: StateT) -> StateT:
        """Register a new state type."""
        if not isinstance(state, State):
            raise TypeError("State must be an instance of State")

        if name in self.states:
            raise ArgumentError(None, f"State '{name}' is already registered")
        self.states[name] = state
        return state

    def load(self):
        """Load all states from disk."""
        logger.info("Loading state...")
        for name, state in self.states.items():
            try:
                if state.changed():
                    logger.warning(f"{name} state has unsaved changes, skipping load")
                    continue
                filepath = self._get_state_path(name)
                state.load(str(filepath))
                logger.info(f"Loaded {name} state")
            except Exception as e:
                logger.warning(f"Failed to load {name} state: {e}")

    async def save(self):
        """Save changed states to disk."""
        changed = False
        async with self.lock:
            logger.info("Saving state...")
            for name, state in self.states.items():
                if state.changed():
                    try:
                        filepath = self._get_state_path(name)
                        state.save(str(filepath))
                        changed = True
                        logger.debug(f"Saved {name} state")
                    except Exception as e:
                        logger.error(f"Failed to save {name} state: {e}")

        if changed:
            logger.info("All states saved")
            self._need_backup = True
        else:
            logger.debug("No changes detected")

    async def backup(self):
        """Create backup of all registered states."""
        if not self._need_backup:
            logger.debug("No changes detected, skipping backup")
            return

        if not self.backup_dir:
            logger.warning("No backup directory configured")
            return
        if self.backup_dir.is_file():
            logger.error("Invalid backup directory")
            return

        logger.info("Creating backup of all states...")

        try:
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            backup_path = self.backup_dir / timestamp
            backup_path.mkdir(parents=True, exist_ok=True)

            for name, state in self.states.items():
                if state.changed():
                    logger.warning(f"{name} state has unsaved changes, skipping backup")
                    continue

                src = self._get_state_path(name)
                if src.exists():
                    dest = backup_path / f"{name}.json"
                    shutil.copy2(src, dest)
                    logger.debug(f"Backed up {name}")

            logger.info(f"All states backed up to {backup_path}")
            self._need_backup = False
        except Exception as e:
            logger.error(f"Failed to create backup: {e}")

    async def restore(self, timestamp: Optional[str] = None):
        """Restore all states from backup."""
        if not self.backup_dir:
            logger.error("No backup directory configured")
            return
        if self.backup_dir.is_file():
            logger.error("Invalid backup directory")
            return
        if not self.backup_dir.exists():
            logger.error("Backup directory does not exist")
            return

        backup_base = self.backup_dir
        if not timestamp:
            backups = [p for p in backup_base.iterdir() if p.is_dir()]
            if not backups:
                logger.error("No backups found")
                return
            backup_path = max(backups, key=lambda p: p.name)
        else:
            backup_path = backup_base / timestamp
            if not backup_path.exists():
                logger.error(f"Backup {timestamp} does not exist")
                return

        logger.info(f"Restoring states from {backup_path.name}...")

        try:
            restored = 0
            for name, state in self.states.items():
                if state.changed():
                    logger.warning(
                        f"{name} state has unsaved changes, skipping restore"
                    )
                    continue
                backup = backup_path / f"{name}.json"
                target = self._get_state_path(name)
                if backup.exists():
                    shutil.copy2(backup, target)
                    restored += 1
                    logger.debug(f"Restored {name}")

            if restored > 0:
                self.load()
                logger.info(f"Restored {restored} states from {backup_path.name}")
            else:
                logger.warning("No states restored")
        except Exception as e:
            logger.error(f"Failed to restore: {e}")

    states: Dict[str, State]
    """Registry of state handlers by name"""

    users: UsersState
    """State handler for user tracking"""

    groups: GroupsState
    """State handler for group tracking"""

    lock: asyncio.Lock
    """Lock for synchronizing state operations"""

    _need_backup: bool
    """Flag indicating if state changes need backup"""



import datetime
from platform import node, release, system
from typing import Optional

from telethon import TelegramClient
from telethon.tl.custom.dialog import Dialog
from telethon.tl.types import (
    Channel,
    Chat,
    ChatPhoto,
    User,
    UserProfilePhoto,
    UserStatusEmpty,
    UserStatusLastMonth,
    UserStatusLastWeek,
    UserStatusOffline,
    UserStatusOnline,
    UserStatusRecently,
)

from ..config import Config
from ..proxy_manager import Proxy
from ..session import CustomSession


def get_client(
    config: Config, proxy: Optional[Proxy], session: Optional[CustomSession]
) -> TelegramClient:
    """
    Create a Telethon TelegramClient from the given configuration.
    """
    from .. import __version__

    params = {
        "session": session or get_session(config),
        "api_id": config.telegram.api.id,  # API ID from my.telegram.org
        "api_hash": config.telegram.api.hash,  # API hash from my.telegram.org
        "connection_retries": 2,  # Number of attempts before failing
        "device_model": node(),  # Real device model
        "system_version": f"{system()} {release()}",  # Real system details
        "app_version": __version__,  # Package version
        "lang_code": "en",  # Language to use for Telegram
        "timeout": 10,  # Timeout in seconds for requests
    }

    if proxy and not proxy.is_direct:
        params.update(proxy.export())

    return TelegramClient(**params)


def get_session(config: Config, create: bool = False) -> CustomSession:
    """
    Load a CustomSession based on the configuration.
    """
    return CustomSession(session_id=config.telegram.session, create=create)


def get_entity_name(
    entity: User | Chat | Channel,
):
    """Get the display name of a Telegram entity."""
    if hasattr(entity, "first_name") and entity.first_name:
        name = entity.first_name
        if hasattr(entity, "last_name") and entity.last_name:
            name += f" {entity.last_name}"
        return name.strip()
    elif hasattr(entity, "title") and entity.title:
        return entity.title
    elif hasattr(entity, "username") and entity.username:
        return f"@{entity.username}"
    else:
        return "Unknown"


def get_status(entity: User):
    """Get the status of a Telegram user."""
    status = entity.status
    if status is None or isinstance(status, UserStatusEmpty):
        return "N/A"
    elif isinstance(status, UserStatusOnline):
        return "Online"
    elif isinstance(status, UserStatusOffline):
        return f"Last seen at {status.was_online.strftime('%Y-%m-%d %H:%M:%S')}"
    elif isinstance(status, UserStatusRecently):
        return "Last seen recently"
    elif isinstance(status, UserStatusLastWeek):
        return "Last seen within the last week"
    elif isinstance(status, UserStatusLastMonth):
        return "Last seen within the last month"
    else:
        return "N/A"


def get_id_from_input_peer(input_peer) -> int:
    """Extract the unique ID from an input peer object."""
    if hasattr(input_peer, "user_id") and input_peer.user_id is not None:
        return input_peer.user_id
    elif hasattr(input_peer, "chat_id") and input_peer.chat_id is not None:
        return input_peer.chat_id
    elif hasattr(input_peer, "channel_id") and input_peer.channel_id is not None:
        return input_peer.channel_id
    else:
        raise ValueError("Input peer does not have a valid ID attribute.")


def has_profile_photo(entity: User | Chat | Channel) -> bool:
    """Check if the entity has a profile photo."""
    photo = getattr(entity, "photo", None)
    if isinstance(photo, (UserProfilePhoto, ChatPhoto)):
        return True
    return False


def _is_dialog_muted(dialog: Dialog) -> bool:
    try:
        if not dialog.dialog.notify_settings.mute_until:
            return False

        return (
            dialog.dialog.notify_settings.mute_until.timestamp()
            > datetime.datetime.now().timestamp()
        )
    except Exception:
        return False



from . import telegram
from .device import get_device_info
from .general import validate_proxy
from .telegram import get_client, get_entity_name, get_session, get_status

__all__ = [
    "telegram",
    "get_client",
    "get_session",
    "get_entity_name",
    "get_status",
    "validate_proxy",
    "get_device_info",
]



import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

from telethon.sessions import SQLiteSession

from .config import Config
from .exceptions import SessionMismatchError, SessionNotFoundError


def get_search_paths(config: Optional[Config] = None) -> List[Path]:
    """Get a list of all session search paths."""
    if config is None:
        config = Config.get_config()

    return [
        Path.cwd(),
        Path.cwd() / "sessions",
        config.base_path,
        config.data_path / "sessions",
        Path.home() / ".cligram" / "sessions" / config.telegram.api.identifier,
    ]


class CustomSession(SQLiteSession):
    """
    Custom Telethon SQLite session with metadata storage and multi-directory search.
    """

    def __init__(self, session_id: Optional[str] = None, create: bool = False):
        """
        Initialize custom session.

        Args:
            session_id: Session identifier/name or full path
        """
        if session_id is None:
            super().__init__(None)
            return

        session_path = Path(session_id)

        # If full path provided, use it directly
        if (
            session_path.suffix == ".session"
            or session_path.is_absolute()
            or os.path.sep in session_id
            or "/" in session_id
        ):
            if not session_path.exists() and not create:
                raise SessionNotFoundError(f"Session file not found: {session_path}")
            super().__init__(str(session_path))
        else:
            # Search in provided paths
            search_paths = get_search_paths()
            found_path = None
            for search_dir in search_paths:
                candidate = search_dir / f"{session_id}.session"
                if candidate.exists():
                    found_path = candidate
                    break

            # If not found, create in last path
            if found_path is None:
                if not create:
                    raise SessionNotFoundError(f"Session file not found: {session_id}")

                last_dir = search_paths[-1]
                last_dir.mkdir(parents=True, exist_ok=True)
                found_path = last_dir / f"{session_id}.session"

            super().__init__(str(found_path))
        self._initialize_metadata_table()

        config = Config.get_config()
        api_id = config.telegram.api.identifier
        session_api_id = self.get_metadata("api_id")
        if session_api_id is None:
            self.set_metadata("api_id", str(api_id))
        elif session_api_id != api_id:
            raise SessionMismatchError(
                "The session was created with a different API ID."
            )

    def _initialize_metadata_table(self):
        """Create metadata table if it doesn't exist."""
        c = self._cursor()
        c.execute(
            """
            CREATE TABLE IF NOT EXISTS metadata (
                key TEXT PRIMARY KEY,
                value TEXT
            )"""
        )
        c.close()

    def set_metadata(self, key: str, value: Any):
        """Store custom metadata."""
        c = self._cursor()
        c.execute(
            "INSERT OR REPLACE INTO metadata (key, value) VALUES (?, ?)",
            (key, str(value)),
        )
        c.close()

    def get_metadata(self, key: str, default: Any = None) -> Optional[str]:
        """Retrieve custom metadata."""
        c = self._cursor()
        c.execute("SELECT value FROM metadata WHERE key = ?", (key,))
        row = c.fetchone()
        c.close()
        return row[0] if row else default

    def get_all_metadata(self) -> Dict[str, str]:
        """Retrieve all metadata as dictionary."""
        c = self._cursor()
        c.execute("SELECT key, value FROM metadata")
        result = {row[0]: row[1] for row in c.fetchall()}
        c.close()
        return result

    def delete_metadata(self, key: str):
        """Delete metadata entry."""
        c = self._cursor()
        c.execute("DELETE FROM metadata WHERE key = ?", (key,))
        c.close()

    @classmethod
    def list_sessions(cls) -> List[str]:
        """List all session files."""
        sessions: Set[Path] = set()
        for search_dir in get_search_paths():
            sessions.update(search_dir.glob("*.session"))
        return [str(s) for s in sessions]



from pathlib import Path

import typer

from ..config import GLOBAL_CONFIG_DIR, Config

app = typer.Typer(help="Configuration management")


@app.command("create")
def create_config(
    path: Path = typer.Argument(
        Path("config.json"), help="Path to create the configuration file at"
    ),
    glb: bool = typer.Option(
        False,
        "--global",
        "-g",
        help="Create global configuration in user home directory",
    ),
    no_interactive: bool = typer.Option(
        False, "--no-interactive", "-n", help="Do not prompt user"
    ),
):
    """Create a default configuration file."""
    if glb:
        path = GLOBAL_CONFIG_DIR / "config.json"

    path = path.resolve()
    if path.exists():
        typer.echo(f"Configuration file already exists at: {path}")
        raise typer.Exit(1)

    def_config = Config()

    if not no_interactive:
        typer.echo("Creating default configuration. Press Enter to accept defaults.")
        api_id = typer.prompt(
            "Telegram API ID", default=str(def_config.telegram.api.id)
        )
        api_hash = typer.prompt(
            "Telegram API Hash", default=def_config.telegram.api.hash
        )
        def_config.telegram.api.id = int(api_id)
        def_config.telegram.api.hash = api_hash

    def_config.save(path)
    typer.echo(f"Created default configuration file at: {path}")


@app.command("list")
def list_configs(
    ctx: typer.Context,
):
    """List all available configuration files."""
    config: Config = ctx.obj["g_load_config"]()
    flatted = Config._flatten_dict(config.to_dict())
    for key, value in flatted.items():
        typer.echo(f"{key}={value}")
    raise typer.Exit()


@app.command("get")
def get_config(
    ctx: typer.Context, key: str = typer.Argument(help="Configuration key to retrieve")
):
    """Get a specific configuration value."""
    config: Config = ctx.obj["g_load_config"]()
    value = config.get_nested_value(key)
    if value is not None:
        typer.echo(f"{key}={value}")
    else:
        typer.echo(f"Configuration key not found: {key}")
    raise typer.Exit()


@app.command("set")
def set_config(
    ctx: typer.Context,
    key: str = typer.Argument(help="Configuration key to set"),
    value: str = typer.Argument(help="Configuration value to set"),
):
    """Set a specific configuration value."""
    config: Config = ctx.obj["g_load_config"]()
    parsed_value = config._parse_value(value)
    config._set_nested_value(key, parsed_value)
    config.save()
    nvalue = config.get_nested_value(key)
    typer.echo(f"{key}={nvalue}")



import logging
from typing import Callable, Coroutine, Optional

from rich.style import Style
from telethon import TelegramClient
from telethon.tl import functions
from telethon.tl.types import User

from .. import (
    Application,
    CustomSession,
    Proxy,
    ProxyManager,
    SessionNotFoundError,
    utils,
)
from ..exceptions import NoWorkingConnectionError

logger: logging.Logger = None  # type: ignore


def setup_logger():
    global logger
    logger = logging.getLogger("cligram.tasks.telegram")


async def setup(
    app: Application,
    callback: Callable[[Application, TelegramClient], Coroutine],
    session: Optional[CustomSession] = None,
    proxy: Optional[Proxy] = None,
):
    """Setup Telegram client."""
    setup_logger()

    try:
        session = session or utils.get_session(app.config)

        if not proxy:
            proxy = await _setup_connection(app)
        _finalize_connection(app, proxy)

        client = await _init_client(app=app, session=session, proxy=proxy)

        try:
            app.check_shutdown()
            await _fetch_account_info(app, client)

            if app.config.telegram.startup.count_unread_messages:
                await _check_unread_messages(app, client)

            await callback(app, client)
        finally:
            app.status.update("Shutting down client...")
            if not client.is_connected():
                logger.warning("Client disconnected unexpectedly")
                app.console.print(
                    "Client disconnected unexpectedly", style=Style(color="yellow")
                )
                return

            await client(functions.account.UpdateStatusRequest(offline=True))

            await client.disconnect()  # type: ignore
            logger.info("Client session closed")

    except SessionNotFoundError as e:
        app.console.print(
            "Session not found:",
            app.config.telegram.session,
            style=Style(color="red"),
        )
        logger.error(f"Session not found: {e}")
    except NoWorkingConnectionError as e:
        app.console.print("No working connection available", style=Style(color="red"))
        logger.error(f"No working connection available: {e}")
    except Exception as e:
        logger.error(f"Unexpected error: {e}")
        raise


async def _setup_connection(app: Application) -> Proxy | None:
    app.status.update("Testing connections...")
    logger.info("Testing connections")
    proxy_manager = ProxyManager.from_config(app.config)
    await proxy_manager.test_proxies(shutdown_event=app.shutdown_event, oneshot=True)
    app.check_shutdown()
    return proxy_manager.current_proxy


def _finalize_connection(app, proxy):
    if utils.validate_proxy(proxy):
        if proxy.is_direct:
            logger.info("Using direct connection")
        else:
            app.console.print(
                f"Using {proxy.type.value} proxy: {proxy.host}:{proxy.port}"
            )
            logger.info(f"Using proxy: [{proxy.type.name}] {proxy.host}:{proxy.port}")
    else:
        logger.error("No working connection available, aborting")
        raise NoWorkingConnectionError("No working connection available")


async def _check_unread_messages(app, client):
    app.status.update("Checking unread messages...")
    logger.debug("Checking for unread messages")
    total_unread = 0
    async for dialog in client.iter_dialogs(limit=50):
        logger.debug(f"Processing {dialog.name} ({dialog.id})")
        try:
            unread = int(getattr(dialog, "unread_count", 0) or 0)
        except Exception:
            unread = 0
        logger.debug(f"Unread count: {unread}")
        if unread <= 0:
            continue
        muted = utils.telegram._is_dialog_muted(dialog)
        logger.debug(f"Is muted: {muted}")
        if muted:
            continue
        total_unread += unread

    if total_unread > 0:
        app.console.print(
            f"You have {total_unread} unread messages",
            style=Style(color="yellow"),
        )
        logger.warning(f"You have {total_unread} unread messages")
    else:
        logger.debug("No unread messages")


async def _fetch_account_info(app, client):
    app.status.update("Fetching account information...")
    logger.debug("Fetching account information")
    me: User = await client.get_me()

    if me.first_name and me.last_name:
        name = f"{me.first_name} {me.last_name}"
    else:
        name = me.first_name

    await client(functions.account.UpdateStatusRequest(offline=False))

    app.console.print(f"Logged in as {name} (ID: {me.id})")
    logger.info(f"Logged in as {name} (ID: {me.id})")

    if app.config.app.verbose:
        logger.debug(f"Account ID: {me.id}")
        logger.debug(f"Full Name: {name}")
        logger.debug(f"Username: {me.username}")
        logger.debug(f"Phone: {me.phone}")


async def _init_client(
    app: Application,
    session: CustomSession,
    proxy: Optional[Proxy] = None,
) -> TelegramClient:
    """Initialize Telegram client."""
    app.status.update("Initializing client...")

    client: TelegramClient = utils.get_client(
        config=app.config, proxy=proxy, session=session
    )

    app.status.update("Logging in...")
    logger.info(f"Logging in with {client.session.filename} session")  # type: ignore

    def _phone_callback():
        app.status.stop()
        return input("Please enter your phone (or bot token): ")

    await client.start(phone=_phone_callback)  # type: ignore
    app.status.start()
    return client



import hashlib
import logging
from logging import FileHandler
from pathlib import Path


class ColoredNameFormatter(logging.Formatter):
    """Formatter that assigns consistent colors to logger names using Rich markup."""

    # Rich color palette for logger names (distinct, readable colors)
    COLORS = [
        "cyan",
        "magenta",
        "green",
        "yellow",
        "blue",
        "bright_cyan",
        "bright_magenta",
        "bright_green",
        "bright_yellow",
        "bright_blue",
        "deep_sky_blue1",
        "medium_purple",
        "spring_green1",
        "gold1",
        "orange1",
    ]

    def __init__(self, fmt=None, datefmt=None, style="%"):
        super().__init__(fmt, datefmt, style)  # type: ignore
        self._color_cache = {}

    def _get_color_for_name(self, name: str) -> str:
        """Get consistent color for a logger name using hash."""
        if name not in self._color_cache:
            # Use hash to consistently map names to colors
            hash_value = int(hashlib.md5(name.encode()).hexdigest(), 16)  # nosec
            color_index = hash_value % len(self.COLORS)
            self._color_cache[name] = self.COLORS[color_index]
        return self._color_cache[name]

    def format(self, record):
        # Color the logger name
        original_name = record.name
        color = self._get_color_for_name(original_name)
        record.name = f"[{color}]{original_name}[/{color}]"

        result = super().format(record)

        # Restore original name
        record.name = original_name
        return result


def setup_logger(verbose: bool = False, log_file: str | Path = "cligram.log"):
    logging.basicConfig(
        level=logging.DEBUG,
    )

    logger = logging.getLogger()
    for h in logger.handlers[:]:
        logger.removeHandler(h)
        h.close()

    # add file handler
    logger.handlers
    log_path = Path(log_file)
    log_path.parent.mkdir(parents=True, exist_ok=True)
    fh = FileHandler(log_path, encoding="utf-8")
    fh.setLevel(logging.DEBUG)
    fh_formatter = logging.Formatter("%(asctime)s %(levelname)s %(name)s: %(message)s")
    fh.setFormatter(fh_formatter)
    logger.addHandler(fh)

    if verbose:
        # Add rich logging for console output
        from rich.logging import RichHandler

        console_handler = RichHandler(
            markup=True,
            omit_repeated_times=False,
        )
        console_handler.setLevel(logging.DEBUG)
        console_formatter = ColoredNameFormatter("%(name)s: %(message)s")
        console_handler.setFormatter(console_formatter)
        logger.addHandler(console_handler)



import asyncio
import base64
import re
import time
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Tuple

import socks
from telethon import TelegramClient, sessions
from telethon.network import ConnectionTcpMTProxyRandomizedIntermediate

from .config import Config


class ProxyType(Enum):
    """
    Type of proxy connection supported by the application.

    Variants:
        MTPROTO: Telegram's native MTProto proxy protocol
        SOCKS5: SOCKS5 proxy protocol with optional authentication
        DIRECT: Direct connection without proxy
    """

    MTPROTO = "mtproto"
    SOCKS5 = "socks5"
    DIRECT = "direct"


@dataclass
class Proxy:
    """
    Proxy connection configuration.

    Supports both MTProto and SOCKS5 protocols with their respective
    authentication methods. For MTProto, requires secret key; for SOCKS5,
    supports optional username/password authentication.
    """

    url: str
    """Original proxy URL string"""

    type: ProxyType
    """Type of proxy protocol to use"""

    host: str
    """Proxy server hostname or IP address"""

    port: int
    """Proxy server port number"""

    secret: Optional[str] = None
    """MTProto secret key (hex or base64 encoded)"""

    username: Optional[str] = None
    """SOCKS5 authentication username"""

    password: Optional[str] = None
    """SOCKS5 authentication password"""

    @property
    def is_direct(self) -> bool:
        """Check if the proxy is a direct connection."""
        return self.type == ProxyType.DIRECT

    def export(self):
        """
        Export proxy configuration for Telethon client.

        Returns:
            dict: Client-compatible proxy configuration parameters
        """
        params = {}

        if self.type == ProxyType.MTPROTO:
            params["connection"] = ConnectionTcpMTProxyRandomizedIntermediate
            params["proxy"] = (self.host, self.port, self.secret)
        elif self.type == ProxyType.SOCKS5:
            params["proxy"] = (
                socks.SOCKS5,
                self.host,
                self.port,
                True,
                self.username,
                self.password,
            )

        return params

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Proxy):
            return False
        if self.type == ProxyType.DIRECT and other.type == ProxyType.DIRECT:
            return True
        return self.url == other.url

    def __hash__(self) -> int:
        if self.type == ProxyType.DIRECT:
            return hash("direct_connection")
        return hash(self.url)


class ProxyManager:
    """
    Manages proxy connections and testing.

    Features:
    - Parse multiple proxy URL formats (MTProto, SOCKS5)
    - Connection testing with timeouts
    - Automatic proxy selection
    - Proxy failover support
    """

    @classmethod
    def from_config(
        cls, config: Config, exclude_direct: bool = False
    ) -> "ProxyManager":
        """
        Create ProxyManager instance from application config.

        Args:
            config: Application configuration object

        Returns:
            ProxyManager instance
        """
        proxy_manager = cls()
        if config.telegram.connection.direct and not exclude_direct:
            proxy_manager.add_direct_proxy()
        for proxy in config.telegram.connection.proxies:
            proxy_manager.add_proxy(proxy)
        return proxy_manager

    def __init__(self):
        self.proxies: List[Proxy] = []
        self.current_proxy: Optional[Proxy] = None

    proxies: List[Proxy]
    """List of configured proxy connections"""

    current_proxy: Optional[Proxy]
    """Currently selected working proxy"""

    def add_proxy(self, proxy_url: str) -> Optional[Proxy]:
        """
        Add new proxy from URL string.

        Args:
            proxy_url: URL string in supported format

        Returns:
            Configured Proxy instance if parsing successful
        """
        proxy = self._parse_proxy_url(proxy_url)
        if proxy and proxy not in self.proxies:
            self.proxies.append(proxy)
        return proxy

    def add_direct_proxy(self):
        """
        Add direct connection (no proxy) to the manager.
        """
        direct_proxy = Proxy(
            url="",
            type=ProxyType.DIRECT,
            host="",
            port=0,
        )
        if direct_proxy not in self.proxies:
            self.proxies.append(direct_proxy)

    def _decode_secret(self, secret: str) -> str:
        """
        Decode MTProto proxy secret from base64.

        Args:
            secret: Base64 encoded secret string

        Returns:
            Decoded hex string or original if decoding fails
        """
        secret += "=" * ((4 - len(secret) % 4) % 4)
        try:
            return base64.b64decode(secret).hex()
        except Exception:
            return secret

    def _parse_proxy_url(self, proxy_url: str) -> Optional[Proxy]:
        """
        Parse proxy URL into proxy configuration.

        Supports formats:
        - mtproto://<secret>@<host>:<port>
        - tg://proxy?server=<host>&port=<port>&secret=<secret>
        - socks5://[<user>:<pass>@]<host>:<port>

        Args:
            proxy_url: Proxy URL string

        Returns:
            Proxy configuration object or None if parsing fails
        """
        mtproto_match = re.match(r"mtproto://([^@]+)@([^:]+):(\d+)", proxy_url)
        mtproto_tg_match = re.match(
            r"(?:tg|https?://t\.me)/proxy\?server=([^&]+)&port=(\d+)&secret=([^&]+)",
            proxy_url,
        )
        socks5_match = re.match(
            r"socks5://(?:([^:]+):([^@]+)@)?([^:]+):(\d+)", proxy_url
        )

        if mtproto_match:
            secret, host, port = mtproto_match.groups()
            return Proxy(
                type=ProxyType.MTPROTO,
                host=host,
                port=int(port),
                secret=self._decode_secret(secret),
                url=proxy_url,
            )
        elif mtproto_tg_match:
            host, port, secret = mtproto_tg_match.groups()
            return Proxy(
                type=ProxyType.MTPROTO,
                host=host,
                port=int(port),
                secret=self._decode_secret(secret),
                url=proxy_url,
            )
        elif socks5_match:
            username, password, host, port = socks5_match.groups()
            return Proxy(
                type=ProxyType.SOCKS5,
                host=host,
                port=int(port),
                username=username,
                password=password,
                url=proxy_url,
            )
        return None

    async def test_proxies(
        self,
        filter: Optional[ProxyType] = None,
        exclusion: List[Proxy] = [],
        shutdown_event: Optional[asyncio.Event] = None,
        timeout: float = 30.0,
        oneshot: bool = False,
    ) -> List["ProxyTestResult"]:

        candidates = [
            p
            for p in self.proxies
            if (not filter or p.type == filter) and p not in exclusion
        ]

        if not candidates:
            return []

        results = await test_proxies(
            candidates, timeout=timeout, shutdown_event=shutdown_event, oneshot=oneshot
        )

        # Return first working proxy
        for result in results:
            if result.success:
                self.current_proxy = result.proxy

        return results


MT_PING_TAG = b"\xee\xee\xee\xee"


@dataclass
class ProxyTestResult:
    """Result of a proxy test including latency and status."""

    proxy: "Proxy"  # Forward reference
    success: bool
    latency: Optional[float] = None
    error: Optional[str] = None

    @property
    def score(self) -> float:
        """Calculate proxy score (lower is better)."""
        if not self.success:
            return float("inf")
        return self.latency or float("inf")

    @property
    def is_good(self) -> bool:
        """Check if the proxy test was successful."""
        return self.success and (self.latency is not None and self.latency < 1000)


async def ping_mtproto(
    proxy: "Proxy", timeout: float = 30.0
) -> Tuple[bool, Optional[float], Optional[str]]:
    """Test MTProto proxy using actual Telegram connection."""
    return await _test_telegram_connection(
        timeout=timeout,
        connection=ConnectionTcpMTProxyRandomizedIntermediate,
        proxy=(proxy.host, proxy.port, proxy.secret),
    )


async def ping_socks5(
    proxy: "Proxy", timeout: float = 5.0
) -> Tuple[bool, Optional[float], Optional[str]]:
    """Test SOCKS5 proxy with timeout and error reporting."""
    start = time.time()
    try:
        reader, writer = await asyncio.wait_for(
            asyncio.open_connection(proxy.host, proxy.port), timeout
        )
        # Initial greeting
        writer.write(b"\x05\x01\x00")
        await writer.drain()
        resp = await asyncio.wait_for(reader.readexactly(2), timeout)
        if resp != b"\x05\x00":
            return False, None, "SOCKS5 no-auth refused"

        writer.write(b"\x05\x01\x00\x01\x00\x00\x00\x00\x00\x00")
        await writer.drain()

        header = await asyncio.wait_for(reader.readexactly(4), timeout)
        if header[0] != 5 or header[1] != 0:
            return False, None, f"SOCKS5 connect failed (REP={header[1]})"

        # Clean up remaining data
        atyp = header[3]
        try:
            if atyp == 1:
                await reader.readexactly(4 + 2)
            elif atyp == 3:
                await reader.readexactly(ord(await reader.readexactly(1)) + 2)
            elif atyp == 4:
                await reader.readexactly(16 + 2)
            else:
                return False, None, f"Unknown address type: {atyp}"
        except Exception:
            pass  # Ignore errors in cleanup

        writer.close()
        try:
            await writer.wait_closed()
        except Exception:
            pass  # Ignore close errors
        return True, (time.time() - start) * 1000, None
    except asyncio.TimeoutError:
        return False, None, "timed out"
    except Exception as e:
        return False, None, str(e)


async def ping_direct(
    proxy: "Proxy", timeout: float = 30.0
) -> Tuple[bool, Optional[float], Optional[str]]:
    """Test direct connection to Telegram servers."""
    return await _test_telegram_connection(timeout=timeout, proxy=proxy)


async def _test_telegram_connection(
    timeout: float = 30.0,
    connection=None,
    proxy=None,
) -> Tuple[bool, Optional[float], Optional[str]]:
    """Helper function to test Telegram connection with or without proxy."""
    start = time.time()
    client = None

    try:
        # Create temporary client for testing
        kwargs = {
            "timeout": timeout,
            "auto_reconnect": False,
            "connection_retries": 1,
            "retry_delay": 1,
        }

        if connection:
            kwargs["connection"] = connection
        if not isinstance(proxy, Proxy) or proxy.type != ProxyType.DIRECT:
            kwargs["proxy"] = proxy

        client = TelegramClient(
            None,  # Memory session  # type: ignore
            1,  # Dummy API ID
            "0" * 32,  # Dummy API hash
            **kwargs,
        )

        # Test basic connection
        await asyncio.wait_for(client.connect(), timeout)
        success = True
        error = None

    except asyncio.TimeoutError:
        success = False
        error = "timed out"
    except ConnectionError as e:
        if "Invalid DC" in str(e):
            success = True
            error = None
        else:
            success = False
            error = str(e)
    except Exception as e:
        success = False
        error = str(e)
    finally:
        try:
            if isinstance(client, TelegramClient):
                if (
                    isinstance(proxy, Proxy)
                    and proxy.type == ProxyType.DIRECT
                    and isinstance(client.session, sessions.Session)
                ):
                    proxy.host = client.session.server_address
                    proxy.port = client.session.port
                    proxy.url = f"dc:{client.session.dc_id}"
                if client.is_connected():
                    await client.disconnect()  # type: ignore
        except Exception:
            pass

    latency = (time.time() - start) * 1000 if success else None
    return success, latency, error


async def test_proxy(
    proxy: "Proxy", timeout: float = 5.0, shutdown_event: Optional[asyncio.Event] = None
) -> ProxyTestResult:
    """Test a proxy with timeout and shutdown support."""
    if shutdown_event and shutdown_event.is_set():
        return ProxyTestResult(proxy, False, error="Test cancelled")

    if proxy.type == ProxyType.MTPROTO:
        test_func = ping_mtproto
        success, latency, error = await test_func(proxy, timeout)
    elif proxy.type == ProxyType.SOCKS5:
        test_func = ping_socks5
        success, latency, error = await test_func(proxy, timeout)
    elif proxy.type == ProxyType.DIRECT:
        success, latency, error = await ping_direct(proxy, timeout)
    else:
        return ProxyTestResult(proxy, False, error="Unknown proxy type")

    return ProxyTestResult(proxy, success, latency, error)


async def test_proxies(
    proxies: List["Proxy"],
    timeout: float = 30.0,
    shutdown_event: Optional[asyncio.Event] = None,
    oneshot: bool = False,
) -> List[ProxyTestResult]:
    """Test multiple proxies concurrently and return sorted results."""
    if not proxies:
        return []

    # Create tasks explicitly using asyncio.create_task
    tasks = [
        asyncio.create_task(test_proxy(p, timeout, shutdown_event)) for p in proxies
    ]
    results = []

    try:
        # Use as_completed to get results as they finish
        for coro in asyncio.as_completed(tasks):
            result = await coro
            results.append(result)

            if oneshot or (shutdown_event and shutdown_event.is_set()):
                break
    except Exception:
        pass
    finally:
        # Cancel any remaining tasks
        for task in tasks:
            if not task.done():
                task.cancel()
                try:
                    await task
                except asyncio.CancelledError:
                    pass

    # Sort by score (successful proxies first, then by latency)
    results.sort(key=lambda r: r.score)
    return results



from dataclasses import dataclass
from enum import Enum


class OperationSystem(Enum):
    UNKNOWN = "Unknown"
    WINDOWS = "Windows"
    LINUX = "Linux"
    ANDROID = "Android"


class Architecture(Enum):
    UNKNOWN = "unknown"
    X86 = "x86"
    X64 = "x64"
    ARM = "arm"
    ARM64 = "arm64"


@dataclass
class DeviceInfo:
    os: OperationSystem
    architecture: Architecture
    os_title: str
    version: str
    model: str


def get_device_info() -> DeviceInfo:
    import platform

    system = platform.system()
    architecture = get_architecture()
    model = platform.node()
    version = platform.release()

    if system == "Windows":
        os = OperationSystem.WINDOWS
        mb_model = _windows_get_motherboard_model_registry()
        if mb_model:
            model = mb_model
        version = platform.win32_ver()[0]
    elif system == "Linux":
        if "android" in platform.platform().lower():
            os = OperationSystem.ANDROID
        else:
            os = OperationSystem.LINUX
    else:
        os = OperationSystem.UNKNOWN

    os_title = f"{os.value} {version} {architecture.value}"

    return DeviceInfo(
        os=os,
        architecture=architecture,
        os_title=os_title,
        version=version,
        model=model,
    )


def get_architecture() -> Architecture:
    import platform

    machine = platform.machine().lower()

    # x64/AMD64
    if machine in ("amd64", "x86_64", "x64"):
        return Architecture.X64
    # ARM64
    elif machine in ("arm64", "aarch64", "armv8", "armv8l", "aarch64_be"):
        return Architecture.ARM64
    # ARM (32-bit)
    elif machine.startswith("arm") or machine in ("armv7l", "armv6l", "armv5l"):
        return Architecture.ARM
    # x86
    elif machine in ("i386", "i686", "x86", "i86pc"):
        return Architecture.X86
    else:
        return Architecture.UNKNOWN


def _windows_get_motherboard_model_registry() -> str | None:
    import winreg

    try:
        key = winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE, r"HARDWARE\DESCRIPTION\System\BIOS"
        )
        value, _ = winreg.QueryValueEx(key, "SystemProductName")
        winreg.CloseKey(key)
        return value
    except Exception:
        return None



import asyncio
from typing import List

import typer
from rich.console import Console
from rich.style import Style
from rich.table import Table

from ..config import Config
from ..proxy_manager import Proxy, ProxyManager, ProxyTestResult

app = typer.Typer(
    help="Manage proxy settings and test proxy connectivity",
    add_completion=False,
)


def _get_proxy_title(proxy: Proxy, use_url) -> str:
    if use_url:
        return proxy.url
    else:
        return f"[{proxy.type.value}] {_get_proxy_host(proxy, use_url=False)}"


def _get_proxy_host(proxy: Proxy, use_url: bool) -> str:
    if proxy.is_direct and not proxy.url:
        return "direct"

    if use_url:
        return proxy.url
    else:
        return f"{proxy.host}:{proxy.port}"


async def run_tests(
    proxy_manager: ProxyManager,
    shutdown_event: asyncio.Event | None = None,
    use_url: bool = False,
    timeout: float = 30.0,
):
    if shutdown_event is None:
        shutdown_event = asyncio.Event()

    typer.echo(f"Testing {len(proxy_manager.proxies)} proxies...")

    results = []

    try:
        results = await proxy_manager.test_proxies(
            shutdown_event=shutdown_event,
            timeout=timeout,
        )

        con = Console()
        table = create_console_table(con, use_url)
        table.add_column("Status", justify="center", vertical="middle", max_width=15)

        for i, result in enumerate(results):
            status = f"{result.latency:.0f}ms" if result.success else result.error

            style = Style(
                color=(
                    "red" if not result.success else "green" if result.is_good else None
                )
            )

            table.add_row(
                str(i + 1),
                result.proxy.type.value,
                _get_proxy_host(result.proxy, use_url=use_url),
                status,
                style=style,
            )
        con.print(table)
    finally:
        shutdown_event.set()

    return results


def create_console_table(con: Console, use_url: bool):
    table = Table(show_header=True)
    table.add_column("#", justify="center", vertical="middle")
    table.add_column("Type", justify="center", vertical="middle")
    table.add_column("URL" if use_url else "Host", overflow="fold", vertical="middle")
    return table


@app.command("add")
def add_proxy(
    ctx: typer.Context,
    url: List[str] = typer.Argument(help="Proxy URL(s) (mtproto:// or socks5://)"),
    skip_test: bool = typer.Option(
        False, "--skip-test", help="Skip testing the proxy before adding"
    ),
):
    """
    Add a new proxy to the configuration.
    """
    config: Config = ctx.obj["g_load_config"]()
    proxy_manager = ProxyManager()

    for proxy_url in url:
        proxy_manager.add_proxy(proxy_url)

    if not proxy_manager.proxies:
        typer.echo("Failed to add proxy. Please check the URL format.")
        raise typer.Exit(code=1)

    pending: List[Proxy] = []
    if not skip_test:
        results: list[ProxyTestResult] = asyncio.run(
            run_tests(proxy_manager, shutdown_event=None, use_url=False)
        )
        pending = [result.proxy for result in results if result.success]
    else:
        pending = proxy_manager.proxies

    c = 0
    for proxy in pending:
        if proxy.url not in config.telegram.connection.proxies:
            config.telegram.connection.proxies.append(proxy.url)
            c += 1
    if c > 0:
        typer.echo(f"Added {c} new proxy(s) to the configuration.")
        config.save()
    else:
        typer.echo("No new proxies were added to the configuration.")
        raise typer.Exit(code=1)


@app.command("list")
def list_proxies(
    ctx: typer.Context,
    show_url: bool = typer.Option(
        False, "--show-url", help="Show full proxy URL in the output"
    ),
):
    """
    List all configured proxies.
    """
    config: Config = ctx.obj["g_load_config"]()
    proxy_manager = ProxyManager.from_config(config)

    if not proxy_manager.proxies:
        typer.echo("No proxies configured.")
        raise typer.Exit(code=1)

    typer.echo("Configured Proxies:")
    con = Console()
    table = create_console_table(con, use_url=show_url)
    for i, proxy in enumerate(proxy_manager.proxies):
        table.add_row(
            str(i + 1),
            proxy.type.value,
            _get_proxy_host(proxy, use_url=show_url),
        )
    con.print(table)


@app.command("test")
def test_proxies(
    ctx: typer.Context,
    show_url: bool = typer.Option(
        False, "--show-url", help="Show full proxy URL in the output"
    ),
    timeout: float = typer.Option(30.0, "--timeout", "-t", help="Timeout in seconds"),
):
    """
    Test all configured proxies and report their status.
    """
    config: Config = ctx.obj["g_load_config"]()
    proxy_manager = ProxyManager.from_config(config)

    asyncio.run(
        run_tests(proxy_manager, shutdown_event=None, use_url=show_url, timeout=timeout)
    )


@app.command("remove")
def remove_proxy(
    ctx: typer.Context,
    url: List[str] = typer.Argument(help="Proxy URL(s) (mtproto:// or socks5://)"),
    all: bool = typer.Option(
        False, "--all", "-a", help="Remove all configured proxies"
    ),
    unreachable: bool = typer.Option(
        False, "--unreachable", "-u", help="Remove all unreachable proxies"
    ),
):
    """
    Remove a proxy from the configuration.
    """
    config: Config = ctx.obj["g_load_config"]()

    unreachable_proxies: List[str] = []

    if all:
        config.telegram.connection.proxies.clear()
        typer.echo("Removed all proxies from the configuration.")
        config.save()
        raise typer.Exit()

    if unreachable:
        proxy_manager = ProxyManager.from_config(config, exclude_direct=True)
        results: list[ProxyTestResult] = asyncio.run(
            run_tests(proxy_manager, shutdown_event=None, use_url=False)
        )
        unreachable_proxies = [
            result.proxy.url for result in results if not result.success
        ]
        typer.echo(f"Found {len(unreachable_proxies)} unreachable proxy(s).")

    c = 0
    target_urls = set(url + unreachable_proxies)
    for proxy_url in target_urls:
        if proxy_url == "direct":
            typer.echo("You must disable direct connection manually in the config.")
            continue
        if proxy_url in config.telegram.connection.proxies:
            config.telegram.connection.proxies.remove(proxy_url)
            c += 1
    if c > 0:
        typer.echo(f"Removed {c} proxy(s) from the configuration.")
        config.save()
    else:
        typer.echo("No matching proxies found.")
        raise typer.Exit(code=1)



from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("cligram")
except PackageNotFoundError:
    # Package not installed, try reading from pyproject.toml
    import tomllib
    from pathlib import Path

    try:
        pyproject_path = Path(__file__).parent.parent.parent / "pyproject.toml"
        with open(pyproject_path, "rb") as f:
            pyproject = tomllib.load(f)
        __version__ = pyproject["project"]["version"]
    except (FileNotFoundError, KeyError):
        raise RuntimeError("Cannot determine the version of cligram.")



import argparse
import asyncio
import logging
import shlex
import sys
import time
import traceback
from contextlib import redirect_stderr, redirect_stdout
from dataclasses import dataclass
from io import StringIO
from typing import Any, Awaitable, Callable, List, Optional

from rich.control import Control, ControlType  # type: ignore
from rich.table import Table
from telethon import TelegramClient, events, functions, hints
from telethon.tl.types import Channel, Message, TypeInputPeer, User, Username

from .. import Application, InteractiveMode, utils
from . import telegram

logger: logging.Logger = None  # type: ignore


def setup_logger():
    global logger
    logger = logging.getLogger("cligram.tasks.interactive")


class InputHandler:
    """Handle async input without blocking output or breaking mid-type input."""

    def __init__(self, context: "Context"):
        self.context = context
        self.input_lock = asyncio.Lock()
        self.input_queue: asyncio.Queue[str] = asyncio.Queue()

        self._prompt_text: str | None = None

    def print_prompt(self):
        """Print the input prompt."""
        prefix = self._prompt_text or ""
        self.context.console.print(f"{prefix}> ", end="", highlight=False)

    async def read_input(self) -> str:
        """Read a line of input asynchronously."""
        return await self.input_queue.get()

    async def start(self):
        """Start the input reader task."""
        asyncio.create_task(self._input_reader())

    async def _input_reader(self):
        """Background task to read stdin."""
        loop = asyncio.get_event_loop()
        while True:
            try:
                # Read input in executor to avoid blocking
                line = await loop.run_in_executor(None, sys.stdin.readline)
                if not line:
                    break
                await self.input_queue.put(line.rstrip("\n"))
            except Exception as e:
                logger.error(f"Error reading input: {e}")
                break

    async def safe_print(self, *args, **kwargs):
        """Print output while preserving input prompt."""
        async with self.input_lock:
            # Clear current line
            self.context.console.control(Control(ControlType.CARRIAGE_RETURN))
            self.context.console.print(" " * self.context.console.width, end="")
            self.context.console.control(Control(ControlType.CARRIAGE_RETURN))

            # Print the message
            self.context.console.print(*args, **kwargs)


@dataclass
class Command:
    name: str
    """The name of the command."""

    aliases: List[str]
    """The list of aliases for the command."""

    description: str
    """The description of the command."""

    parser: argparse.ArgumentParser
    """The argument parser for the command."""

    handler: Callable[["CommandHandler", argparse.Namespace], Awaitable[None]]
    """The function that handles the command."""


class CommandHandler:
    """Handle user commands in interactive mode."""

    def __init__(self, context: "Context"):
        self.context = context
        self.selected_entity: Optional[TypeInputPeer] = None
        self.commands: dict[str, Command] = {}

        self._add_select()
        self._add_resolve()
        self._add_send()

    def _add_select(self):
        selectParser = argparse.ArgumentParser(
            prog="select", description="Select a Telegram entity", exit_on_error=False
        )
        selectParser.add_argument(
            "entity",
            nargs="?",
            default=None,
            type=str,
            help="Username or ID of the entity to select",
        )
        selectParser.add_argument(
            "-c", "--clear", action="store_true", help="Clear the current selection"
        )

        self.add_command(
            Command(
                name="select",
                aliases=[],
                description="Select a Telegram entity by username or ID",
                parser=selectParser,
                handler=self.cmd_select,
            )
        )

    def _add_resolve(self):
        resolveParser = argparse.ArgumentParser(
            prog="resolve", description="Resolve a Telegram entity", exit_on_error=False
        )
        resolveParser.add_argument(
            "-e",
            "--entity",
            default=None,
            type=str,
            help="Username or ID of the entity to resolve, if not provided, will use the selected entity",
        )
        resolveParser.add_argument(
            "-a", "--all", action="store_true", help="Show full data of the entity"
        )

        self.add_command(
            Command(
                name="resolve",
                aliases=[],
                description="Resolve a Telegram entity by username or ID",
                parser=resolveParser,
                handler=self.cmd_resolve,
            )
        )

    def _add_send(self):
        sendParser = argparse.ArgumentParser(
            prog="send",
            description="Send a message to the selected entity",
            exit_on_error=False,
        )
        sendParser.add_argument(
            "message",
            type=str,
            help="The message text to send",
        )

        sendParser.add_argument(
            "-e",
            "--entity",
            default=None,
            type=str,
            help="Username or ID of the entity to send the message to, if not provided, will use the selected entity",
        )

        self.add_command(
            Command(
                name="send",
                aliases=[],
                description="Send a message to the selected entity",
                parser=sendParser,
                handler=self.cmd_send,
            )
        )

    def add_command(self, command: Command):
        """Add a new command."""
        if command.name in self.commands:
            raise ValueError(f"Command {command.name} already exists.")

        if not isinstance(command.parser, argparse.ArgumentParser):
            raise TypeError(
                "Command parser must be an instance of argparse.ArgumentParser."
            )

        if command.parser.exit_on_error:
            raise ValueError("Command parser must not exit on error.")

        if not callable(command.handler):
            raise TypeError("Command handler must be callable.")

        if not asyncio.iscoroutinefunction(command.handler):
            raise TypeError("Command handler must be an async function.")

        self.commands[command.name] = command

    async def handle_command(self, command: str):
        """Handle user commands."""
        parts = shlex.split(command)
        cmd = parts[0]

        if cmd == "help":
            self.context.console.print("[bold]Available Commands:[/bold]")
            table = Table.grid(padding=(0, 5))
            table.add_column("Command")
            table.add_column("Description")

            for command_obj in self.commands.values():
                table.add_row(
                    f"[bold]{command_obj.name}[/bold]", command_obj.description
                )
            self.context.console.print(table)
            self.context.console.print(
                "\nType <command> --help for more details on a command."
            )
        elif cmd in self.commands:
            command_obj = self.commands[cmd]
            try:
                parsed = command_obj.parser.parse_args(parts[1:])
            except argparse.ArgumentError as e:
                self.context.console.print(f"[red]Argument error:[/red] {e}")
                return
            except SystemExit:
                return
            await command_obj.handler(self, parsed)
        else:
            self.context.console.print(f"[dim]Unknown command: {command}[/dim]")

    async def get_input_entity(self, query: str | int | None):
        """Get input entity from query or selected entity."""
        if query is None:
            if self.selected_entity is None:
                raise ValueError("No entity selected.")
            return self.selected_entity

        try:
            query = int(query)
        except ValueError:
            pass

        return await self.context.client.get_input_entity(query)

    async def cmd_select(self, _, args: argparse.Namespace):
        """Handler for select command."""
        if not isinstance(args, argparse.Namespace):
            self.context.console.print(
                "[red]Invalid arguments for select command.[/red]"
            )
            return

        if args.clear:
            self.selected_entity = None
            self.context.set_prompt(source=InteractiveMode.CLIGRAM, prompt=None)
            self.context.console.print("[green]Selection cleared.[/green]")
            return

        entity_query = args.entity
        if not entity_query:
            self.context.console.print("[red]No entity provided to select.[/red]")
            return

        try:
            ientity = await self.get_input_entity(entity_query)
            id = utils.telegram.get_id_from_input_peer(ientity)
            if id is None:
                raise ValueError("Could not determine entity ID.")
            self.selected_entity = ientity
            self.context.set_prompt(source=InteractiveMode.CLIGRAM, prompt=str(id))
            self.context.console.print(
                f"[green]Entity selected:[/green] {type(ientity).__name__}:{id}"
            )
        except Exception as e:
            self.context.console.print(f"[red]Error selecting entity:[/red] {e}")

    async def cmd_resolve(self, _, args: argparse.Namespace):
        """Handler for resolve command."""
        if not isinstance(args, argparse.Namespace):
            self.context.console.print(
                "[red]Invalid arguments for resolve command.[/red]"
            )
            return

        entity_name = args.entity
        extend = args.all

        try:
            ientity = await self.get_input_entity(entity_name)
            entity = await self.context.client.get_entity(ientity)
            if isinstance(entity, list):
                entity = entity[0]
            self.context.console.print(f"[green]Entity resolved:[/green] {entity.id}")
            table = Table.grid(padding=(0, 5))
            table.add_column("Field")
            table.add_column("Value")

            bot = isinstance(entity, User) and getattr(entity, "bot", False)
            group = isinstance(entity, Channel) and (
                getattr(entity, "megagroup", False)
                or getattr(entity, "gigagroup", False)
            )

            # Show additional info about this entity
            table.add_row("Type", entity.__class__.__name__)
            if isinstance(entity, User):
                table.add_row("Is Bot", str(bot))
            elif isinstance(entity, Channel):
                table.add_row("Is Group", str(group))
            table.add_row("Name", utils.get_entity_name(entity))
            table.add_row(
                "Has Profile Photo", str(utils.telegram.has_profile_photo(entity))
            )

            username = getattr(entity, "username", None)
            usernames: List[Username] = getattr(entity, "usernames", [])
            if username:
                table.add_row("Username", username)
            elif usernames:
                table.add_row("Usernames", "\n".join([u.username for u in usernames]))
            if isinstance(entity, User) and not bot:
                table.add_row("Phone", _tryattr(entity, "phone"))
                table.add_row("Status", utils.get_status(entity))
            if bot:
                table.add_row("Bot Users", _tryattr(entity, "bot_active_users"))
            table.add_row("Is Verified", _tryattr(entity, "verified"))
            table.add_row("Is Scam", _tryattr(entity, "scam"))
            table.add_row("Is Restricted", _tryattr(entity, "restricted"))
            if extend:
                table.add_row("Full Data", entity.stringify())
            self.context.console.print(table)

        except Exception as e:
            self.context.console.print(f"[red]Error resolving entity:[/red] {e}")

    async def cmd_send(self, _, args: argparse.Namespace):
        """Handler for send command."""
        if not isinstance(args, argparse.Namespace):
            self.context.console.print("[red]Invalid arguments for send command.[/red]")
            return

        entity_query = args.entity
        message_text = args.message

        try:
            ientity = await self.get_input_entity(entity_query)
            id = utils.telegram.get_id_from_input_peer(ientity)
            await self.context.client.send_message(ientity, message_text)
            self.context.console.print(f"[green]Message sent to {id}.[/green]")
        except Exception as e:
            self.context.console.print(f"[red]Error sending message:[/red] {e}")


class PythonExecutor:
    """Execute Python code in the interactive session context."""

    def __init__(self, context: "Context"):
        self.context = context
        self.result_history: list[Any] = []
        self.max_history = 100
        self._pending_tasks: dict[int, TaskAwaiter] = {}
        self._task_counter = 0
        self.locals: dict[str, Any] = {
            "__context__": self.context,
            "_": None,  # Last result
            "__results__": self.result_history,  # All results
            "a": self._await_helper,  # Helper to await coroutines
            "__tasks__": self._pending_tasks,  # Pending tasks
        }
        self.globals: dict[str, Any] = {}

    def _get_result(self, index: int = -1):
        """Get result from history by index (default: last result)."""
        if not self.result_history:
            return None
        try:
            return self.result_history[index]
        except IndexError:
            return None

    def _store_result(self, result):
        """Store result in history and update quick access variable."""
        if result is not None:
            self.result_history.append(result)
            if len(self.result_history) > self.max_history:
                self.result_history.pop(0)

            self.locals["_"] = result

    def _await_helper(self, coro):
        """
        Helper function to await a coroutine and get its result.
        Creates a task and returns a TaskAwaiter object that can be used to get the result.

        Usage:
            task = a(client.get_me())
            # Later...
            task.result()  # Get result when done
            task.done()    # Check if done
        """
        if not asyncio.iscoroutine(coro):
            raise TypeError(f"a() requires a coroutine, got {type(coro).__name__}")

        task_id = self._task_counter
        self._task_counter += 1

        task = asyncio.create_task(coro)
        awaiter = TaskAwaiter(task, task_id)
        self._pending_tasks[task_id] = awaiter

        # Clean up when done
        def cleanup(t):
            # Keep in dict for access, but mark as done
            pass

        task.add_done_callback(cleanup)
        return awaiter

    async def execute(self, code: str) -> tuple[bool, str]:
        """
        Execute Python code and return (success, output).

        Args:
            code: Python code to execute

        Returns:
            Tuple of (success: bool, output: str)
        """
        stdout_capture = StringIO()
        stderr_capture = StringIO()

        try:
            # Check if this is an expression or statement
            try:
                compile(code, "<interactive>", "eval")
                is_expression = True
            except SyntaxError:
                is_expression = False

            result = None
            with redirect_stdout(stdout_capture), redirect_stderr(stderr_capture):
                if is_expression:
                    # Evaluate expression and print result
                    result = eval(code, self.globals, self.locals)  # nosec

                    # Handle awaitable results
                    if asyncio.iscoroutine(result) or asyncio.isfuture(result):
                        result = await result

                    # Store the result
                    self._store_result(result)

                    if result is not None:
                        print(repr(result))
                else:
                    # Execute statement
                    compiled = compile(code, "<interactive>", "exec")
                    exec(compiled, self.globals, self.locals)  # nosec

            output = stdout_capture.getvalue()
            errors = stderr_capture.getvalue()

            full_output = ""
            if output:
                full_output += output
            if errors:
                full_output += errors

            return True, full_output.rstrip() if full_output else ""

        except Exception as e:
            error_output = stderr_capture.getvalue()
            if error_output:
                error_output += "\n"
            error_output += "".join(
                traceback.format_exception(type(e), e, e.__traceback__)
            )
            return False, error_output.rstrip()

    async def execute_and_print(self, code: str):
        """Execute code and print the result to console."""
        success, output = await self.execute(code)

        if output:
            await self.context.input_handler.safe_print(output)
        elif not success:
            await self.context.input_handler.safe_print(
                "[red]Execution failed with no output[/red]"
            )

    def add_variable(self, name: str, value):
        """Add a variable to the executor's local scope."""
        self.locals[name] = value

    def get_variable(self, name: str):
        """Get a variable from the executor's local scope."""
        return self.locals.get(name)

    def list_variables(self) -> dict:
        """Get all variables in the executor's local scope."""
        return {k: v for k, v in self.locals.items() if not k.startswith("_")}

    def clear_history(self):
        """Clear result history."""
        self.result_history.clear()
        self.locals["_"] = None


class TaskAwaiter:
    """Wrapper for async tasks to allow easy result access."""

    def __init__(self, task: asyncio.Task, task_id: int):
        self.task = task
        self.task_id = task_id

    def done(self) -> bool:
        """Check if the task is done."""
        return self.task.done()

    def result(self):
        """
        Get the result of the task. Blocks if not done yet.

        Returns:
            The result of the coroutine

        Raises:
            asyncio.TimeoutError: If timeout is reached
            Exception: Any exception raised by the coroutine
        """
        while not self.task.done():
            time.sleep(0.1)

        return self.task.result()

    def exception(self):
        """Get the exception raised by the task, if any."""
        if not self.task.done():
            return None
        return self.task.exception()

    def cancel(self):
        """Cancel the task."""
        return self.task.cancel()

    def __repr__(self):
        status = "done" if self.task.done() else "pending"
        return f"<TaskAwaiter {self.task_id} ({status})>"

    def __await__(self):
        """Allow awaiting the TaskAwaiter directly."""
        return self.task.__await__()


def _tryattr(obj, attr: str):
    value = getattr(obj, attr, None)
    return str(value) if value is not None else "N/A"


class Context:
    """Context for interactive session."""

    @property
    def mode(self) -> InteractiveMode:
        """Get the current interactive mode."""
        return self._mode

    @property
    def print_announcements(self) -> bool:
        """Whether to print announcements."""
        return self._print_announcements

    def __init__(self, app: Application, client: TelegramClient):
        self.app = app
        self.console = app.console
        self.client = client

        self.input_handler = InputHandler(self)
        self.command_handler = CommandHandler(self)
        self.executor = PythonExecutor(self)

        self._mode: InteractiveMode = None  # type: ignore
        self._available_modes = list(InteractiveMode)
        self._input_processor_task: Optional[asyncio.Task] = None
        self._print_announcements: bool = True
        self._cmd_prompt: str | None = None
        self._missed_announcements: asyncio.Queue = asyncio.Queue()

    async def start(self):
        """
        Start the interactive session.
        """
        await self.input_handler.start()
        self._input_processor_task = asyncio.create_task(self._process_input())

        self.console.print("[green]Interactive session started![/green]")
        self.console.print("[dim]Type ! to toggle modes, CTRL+C to exit[/dim]")

        await self.set_mode(self.app.config.interactive.mode)

    async def stop(self):
        """
        Stop the interactive session.
        """

        if self._input_processor_task:
            self._input_processor_task.cancel()
            try:
                await self._input_processor_task
            except asyncio.CancelledError:
                pass

        self.console.print("[green]Interactive session ended[/green]")

    def set_prompt(self, source: InteractiveMode, prompt: str | None):
        """
        Set the input prompt for the given interactive mode.
        """

        if source == InteractiveMode.CLIGRAM:
            self._cmd_prompt = prompt

        self._update_prompt()

    def _update_prompt(self):
        if self._mode == InteractiveMode.CLIGRAM:
            self.input_handler._prompt_text = self._cmd_prompt
        elif self._mode == InteractiveMode.PYTHON:
            self.input_handler._prompt_text = "python"

    async def set_mode(self, mode: InteractiveMode):
        """
        Set the interactive mode.
        """

        self._mode = mode

        if self._mode == InteractiveMode.CLIGRAM:
            self.console.print("[dim]Command mode, type help for commands[/dim]")
            await self.set_print_announcements(True)
        elif self._mode == InteractiveMode.PYTHON:
            self.console.print("[dim]Python executor mode[/dim]")
            await self.set_print_announcements(False)

        self._update_prompt()

    async def set_print_announcements(self, value: bool):
        """
        Enable or disable printing announcements.
        """
        if value == self._print_announcements:
            return
        self._print_announcements = value

        if self._print_announcements:
            self.console.print(
                "[green]Updates announcements enabled, missed announcements will be flushed[/green]"
            )
            await self.flush_missed_announcements()
        else:
            self.console.print(
                "[yellow]Updates announcements disabled, updates still logged[/yellow]"
            )

    async def announce(
        self,
        message: str,
    ):
        """
        Announce a message to the user, if announcements are enabled.
        Otherwise, queue the message for later.

        Args:
            message: The message to announce.
        """
        if not self.print_announcements:
            await self._missed_announcements.put(message)
            return

        await self.input_handler.safe_print(message)

        self.input_handler.print_prompt()

    async def flush_missed_announcements(self):
        """
        Flush all missed announcements.

        Raises:
            RuntimeError: If announcements are disabled.
        """
        if not self.print_announcements:
            raise RuntimeError("Updates announcements are disabled.")

        while not self._missed_announcements.empty():
            message = await self._missed_announcements.get()
            await self.announce(message)

    async def announce_event(
        self,
        event: str,
        entity: hints.Entity,
        text: Optional[str] = None,
    ):
        """
        Announce an event from a Telegram entity.

        Args:
            event: The event name.
            entity: The Telegram entity involved in the event.
            text: Optional additional text to include in the announcement.
        """
        entity_name = utils.get_entity_name(entity)
        formatted_event_message = (
            f"[bold blue]{event}[/bold blue] from [bold green]{entity_name}[/bold green] (ID: {entity.id})"
            + (f"\n{text}" if text else "")
        )

        logger.info(
            f"event '{event}' from {entity_name} (ID: {entity.id})"
            + (f"\n{text}" if text else "")
        )

        await self.announce(formatted_event_message)

    async def _process_input(self):
        self.input_handler.print_prompt()
        while not self.app.shutdown_event.is_set():
            try:
                line = await asyncio.wait_for(
                    self.input_handler.read_input(), timeout=1.0
                )

                if not line.strip():
                    continue

                sline = line.strip()

                if sline == "!":
                    current_index = self._available_modes.index(self._mode)
                    next_index = (current_index + 1) % len(self._available_modes)
                    await self.set_mode(self._available_modes[next_index])
                elif self.mode == InteractiveMode.PYTHON:
                    await self.executor.execute_and_print(sline)
                elif self.mode == InteractiveMode.CLIGRAM:
                    await self.command_handler.handle_command(sline)

                # Print new prompt after command is handled
                self.input_handler.print_prompt()

            except asyncio.TimeoutError:
                continue
            except Exception as e:
                self.console.print(f"[red]Error processing input:[/red] {e}")
                logger.error(f"Error processing input: {e}")


async def main(app: Application):
    """Interactive task."""
    setup_logger()

    app.status.update("Starting interactive session...")
    await telegram.setup(app=app, callback=interactive_callback)


async def interactive_callback(app: Application, client: TelegramClient):
    """Callback for interactive task."""
    app.status.stop()

    context = Context(app, client)
    await context.start()

    @client.on(events.NewMessage)  # type: ignore
    async def handler(event: events.NewMessage.Event):
        msg: Message = event.message
        entity = await client.get_entity(msg.peer_id)
        if isinstance(entity, list):
            entity = entity[0]

        await context.announce_event(
            event="New Message",
            entity=entity,
            text=msg.message,
        )

        # mark message as read
        await client(
            functions.messages.ReadHistoryRequest(
                peer=msg.peer_id, max_id=msg.id  # type: ignore
            )
        )

    # Wait for shutdown event
    await app.shutdown_event.wait()

    await context.stop()

    # Remove event handler
    client.remove_event_handler(handler)

    app.status.start()



def validate_proxy(proxy) -> bool:
    """Validate if the given proxy is valid"""
    from .. import Proxy
    from ..proxy_manager import ProxyType

    if not proxy or not isinstance(proxy, Proxy):
        return False
    if proxy.type not in ProxyType:
        return False
    if not proxy.is_direct and (not proxy.host or not proxy.port):
        return False
    return True



import asyncio
import logging
import platform
import signal
import sys
from typing import Callable, Coroutine, Optional

from rich import get_console
from rich.status import Status

from .config import Config
from .state_manager import StateManager

logger: logging.Logger = None  # type: ignore
app_instance: Optional["Application"] = None
_recv_signals: int = 0


def get_app() -> "Application":
    """
    Retrieve the global application instance.

    Returns:
        Application: The global application instance

    Raises:
        RuntimeError: If application is not initialized
    """
    global app_instance
    if app_instance is None:
        raise RuntimeError("Application instance is not initialized")
    return app_instance


class Application:
    def __init__(self, config: Config):
        global logger

        self.config = config

        logger = logging.getLogger("cligram.app")

        self.state = StateManager(data_dir=self.config.data_path)
        """"State manager for application state persistence."""

        self.shutdown_event: asyncio.Event = asyncio.Event()
        """Event to signal application shutdown."""

        self.console = get_console()
        """Rich console for formatted output."""

        self.status: Status = Status("", console=self.console, spinner="dots")
        """Rich status indicator for CLI feedback."""

    async def shutdown(self, sig=None):
        """
        Handle graceful application shutdown.

        Args:
            sig: Signal that triggered shutdown (SIGTERM/SIGINT)

        Sets shutdown event and allows running operations to complete
        cleanly before terminating.
        """
        global _recv_signals

        if sig:
            _recv_signals += 1
            if _recv_signals >= 3:
                sys.exit(255)
            logger.warning(f"Received exit signal {sig}, count: {_recv_signals}")
            self.console.print(f"[bold red]Received exit signal {sig}[/bold red]")
        if not self.shutdown_event.is_set():
            self.shutdown_event.set()

    async def setup_signal_handlers(self):
        """
        Configure OS signal handlers.

        Handles:
        - SIGTERM for graceful termination
        - SIGINT for keyboard interrupts
        - Platform-specific signal routing
        - Async signal handling
        """
        if platform.system() == "Windows":
            try:
                signal.signal(
                    signal.SIGINT, lambda s, f: asyncio.create_task(self.shutdown(s))
                )
                signal.signal(
                    signal.SIGTERM, lambda s, f: asyncio.create_task(self.shutdown(s))
                )
            except (AttributeError, NotImplementedError):
                logger.warning("Signal handlers not fully supported on Windows")
        else:
            for sig in (signal.SIGTERM, signal.SIGINT):
                try:
                    asyncio.get_event_loop().add_signal_handler(
                        sig, lambda s=sig: asyncio.create_task(self.shutdown(s))  # type: ignore
                    )
                except NotImplementedError:
                    logger.warning(f"Failed to set handler for signal {sig}")

    def check_shutdown(self):
        """
        Check if shutdown has been requested.

        Raises:
            asyncio.CancelledError: If shutdown event is set
        """
        if self.shutdown_event and self.shutdown_event.is_set():
            raise asyncio.CancelledError()

    async def sleep(self):
        """
        Sleep for a configured delay while checking for shutdown.

        Note: this method is not precise and it is an intended behavior

        Raises:
            asyncio.CancelledError: If shutdown is requested during sleep
        """
        remaining = delay = self.config.app.delays.random()
        steps = 0.1
        cur_status = self.status.status

        logger.debug(f"Sleeping for {delay} seconds")
        while True:
            try:
                self.check_shutdown()
                self.status.update(
                    f"[yellow]Sleeping ({round(remaining, 1)})...[/yellow]"
                )
                await asyncio.sleep(steps)
                remaining -= steps
                if remaining <= 0:
                    break
            finally:
                self.status.update(cur_status)

    async def run(self, task: Callable[["Application"], Coroutine]):
        """
        Initialize application and run task.
        """
        from . import __version__

        global app_instance
        if app_instance is not None:
            raise RuntimeError("Application instance is already running")
        app_instance = self

        self.status.update("Starting application...")
        self.status.start()

        self.console.print(f"[bold green]cligram v{__version__}[/bold green]")
        logger.info(f"Starting application v{__version__}")

        self.status.update("Initializing...")
        # Setup platform-specific signal handlers
        await self.setup_signal_handlers()

        logger.debug(f"Loaded configuration: {self.config.path}")

        if self.config.updated:
            self.console.print("[bold yellow]Configuration file updated[/bold yellow]")
            logger.warning("Configuration updated with new fields")

        self.status.update("Loading state...")
        self.state.load()

        try:
            self.status.update("Running task...")
            await task(self)
            logger.info("Execution completed successfully")
        except Exception as e:
            logger.error(f"Error during execution: {e}")
            raise
        finally:
            self.status.update("Shutting down...")
            await self.state.save()
            await self.state.backup()
            logger.info("Shutdown complete")
            self.status.stop()
            app_instance = None

    def start(self, task: Callable[["Application"], Coroutine]):
        """
        Start the application event loop and run the specified task.

        Args:
            task (Callable): Async function representing the main task to run
        """
        if app_instance is not None:
            raise RuntimeError("Application instance is already running")

        if not callable(task):
            raise ValueError("Task must be a callable async function")

        try:
            asyncio.run(self.run(task))
        except (asyncio.CancelledError, KeyboardInterrupt):
            logger.warning("Cancellation requested by user")
            self.console.print(
                "[bold yellow]Application stopped by user request[/bold yellow]"
            )
        except Exception as e:
            logger.fatal(f"Fatal error: {e}", exc_info=True)
            self.console.print(f"[bold red]Fatal error: {e}[/bold red]")
            raise



from telethon import TelegramClient

from .. import Application, CustomSession, SessionNotFoundError, utils
from . import telegram


async def login(app: Application):
    """Login to a new Telegram session."""
    exists = False
    try:
        utils.get_session(app.config, create=False)
        exists = True
    except SessionNotFoundError:
        pass

    if exists:
        app.console.print("[red]Session already exists.[/red]")
        return

    app.status.update("Logging in to Telegram...")
    session: CustomSession = utils.get_session(app.config, create=True)

    await telegram.setup(app=app, callback=login_callback, session=session)


async def login_callback(app: Application, client: TelegramClient):
    """Callback for login task."""
    app.console.print("[green]Logged in successfully![/green]")



import asyncio
import logging
import random
import time
from datetime import datetime
from typing import Dict, List, Optional

from rich.style import Style
from telethon import TelegramClient, errors, events, functions
from telethon.tl.custom.dialog import Dialog
from telethon.tl.types import ChannelParticipantsAdmins, Message, MessageService, User

from . import utils
from .app import Application
from .config import Config, ScanMode
from .exceptions import SessionNotFoundError
from .proxy_manager import ProxyManager
from .state_manager import StateManager

logger: logging.Logger = None  # type: ignore


class TelegramScanner:
    """
    Manages Telegram scanning operations and message forwarding.

    The scanner provides functionality for:
    - Scanning group messages within defined windows
    - Processing users based on configured criteria
    - Forwarding messages with rate limiting
    - Managing multiple proxies for connection resilience
    """

    def __init__(self, app: Application):
        """
        Initialize scanner with configuration and state management.

        Args:
            app: Application instance
        """
        global logger
        logger = logging.getLogger("cligram.scanner")

        self.app: Application = app
        self.config: Config = app.config
        self.state: StateManager = app.state
        self.proxy_manager: ProxyManager = ProxyManager.from_config(app.config)

        self.delay_variation = 10
        self._thresholds = {
            "skip": (
                20 if self.config.scan.mode in (ScanMode.FULL, ScanMode.SEND) else 50
            ),
            "scan": (
                50 if self.config.scan.mode in (ScanMode.FULL, ScanMode.SEND) else 100
            ),
            "process": 20,
        }

        self._cached_ids: Dict[str, str] = {}
        self._entity_cache: Dict[str, User] = {}

    def _get_random_threshold(self, type_: str) -> int:
        """
        Generate a randomized threshold value for rate limiting.

        Args:
            type_: Type of threshold ('skip', 'scan', or 'process')

        Returns:
            Randomized threshold value based on configured base value
        """
        base = self._thresholds[type_]
        return max(
            1, base + random.randint(-self.delay_variation, self.delay_variation)
        )

    async def get_saved_messages(self, client: TelegramClient) -> List[int]:
        """
        Retrieve message IDs from configured source chat.

        Args:
            client: Active Telegram client instance

        Returns:
            List of message IDs from source chat
        """
        limit = self.config.scan.messages.limit
        entity = await client.get_entity(self.config.scan.messages.source)
        logger.debug(f"[SCAN] Using entity {entity.id} for message retrieval")
        msg_ids = []

        if self.config.scan.messages.randomize:
            msgs: List[Message] = await client.get_messages(entity, limit=limit)
            msg_ids = [
                msg.id
                for msg in msgs
                if msg
                and isinstance(msg, Message)
                and not isinstance(msg, MessageService)
                and msg.message
                and not msg.noforwards
            ]
            logger.info(f"[SCAN] Found {len(msg_ids)} message IDs in source")
        else:
            msg_id = self.config.scan.messages.msg_id
            msg = await client.get_messages(entity, ids=msg_id)
            if msg:
                msg_ids.append(msg.id)
                logger.info(f"[SCAN] Using configured message ID {msg.id} from source")
            else:
                raise ValueError(f"Configured message ID {msg_id} not found in source")

        return msg_ids

    async def get_group_id(self, client: TelegramClient, group: str) -> str:
        """
        Resolve a group identifier to its numeric ID.

        Args:
            client: Active Telegram client instance
            group: Group username, URL, or ID string

        Returns:
            String representation of group's numeric ID, or None if resolution fails
        """
        if group in self._cached_ids:
            return self._cached_ids[group]

        try:
            # Check if group is already an ID
            gid_int = int(group)
            logger.debug(f"[GROUP] {group} is already an ID")
            return str(gid_int)
        except ValueError:
            pass

        try:
            entity = await client.get_entity(group)
            if isinstance(entity, list):
                entity = entity[0]
            gid = str(entity.id)
            self._cached_ids[group] = gid
            logger.debug(f"[GROUP] Resolved {group} to ID {gid}")
            return gid
        except Exception as e:
            logger.error(f"[GROUP] Failed to resolve {group}: {e}")
            raise

    async def human_delay(self, shutdown_event: Optional[asyncio.Event] = None):
        """
        Implement a configurable delay with random variation.

        Args:
            shutdown_event: Optional event to interrupt delay

        The delay duration is chosen based on configured normal and long break periods,
        with long break periods occurring based on probability settings.
        """
        delay = self.config.app.delays.random()
        logger.debug(f"[DELAY] Sleeping for {delay:.2f} seconds")

        # Sleep with cancellation support
        try:
            if shutdown_event:
                await asyncio.wait_for(shutdown_event.wait(), timeout=delay)
                logger.debug("[DELAY] Interrupted by shutdown")
                return
            else:
                await asyncio.sleep(delay)
        except asyncio.TimeoutError:
            pass  # Normal delay completion

    async def _resolve_user(
        self, client: TelegramClient, username: str
    ) -> Optional[User]:
        """
        Resolve username to Telegram User entity with caching.

        Args:
            client: Active Telegram client instance
            username: Username to resolve (without @ symbol)

        Returns:
            User entity if resolution successful, None otherwise
        """
        if username in self._entity_cache:
            logger.debug(f"[USER] Using cached entity for @{username}")
            return self._entity_cache[username]

        try:
            entity = await client.get_entity(f"@{username}")
            if not isinstance(entity, User):
                return None
            self._entity_cache[username] = entity
            logger.debug(f"[USER] Resolved @{username} to User ID {entity.id}")
            return entity
        except Exception as e:
            logger.error(f"[USER] Failed to resolve @{username}: {e}")
            return None

    async def process_user(
        self, client: TelegramClient, user: User, msg_ids: List[int]
    ) -> bool:
        """
        Process a user according to current operation mode.

        Args:
            client: Active Telegram client instance
            user: User entity to process
            msg_ids: List of message IDs available for forwarding

        Returns:
            True if user was successfully processed, False if skipped

        Handles scanning, eligibility checking, and message forwarding based on
        configured operation mode.
        """
        uid = user.id
        username = getattr(user, "username", None)

        try:
            # For scan/send modes, require username
            if self.config.scan.mode != ScanMode.FULL and not username:
                logger.debug(f"[SKIP] User {uid} has no username")
                return False

            if uid in self.state.users.messaged:
                logger.debug(f"[SKIP] User {username or uid} already messaged")
                return False

            if self.config.scan.mode == ScanMode.SCAN:
                if username in self.state.users.eligible:
                    logger.debug(f"[SKIP] User {username} already eligible")
                    return False
                self.state.users.eligible.add(username)  # type: ignore
                logger.info(f"[DONE] Added username {username} to eligible list")
                return True

            if self.config.scan.mode in (ScanMode.FULL, ScanMode.SEND):
                msg_id = random.choice(msg_ids)
                if not self.config.scan.test:
                    logger.info(
                        f"[SEND] Forwarding message {msg_id} to user {username or uid}"
                    )
                    await client.forward_messages(
                        user, msg_id, from_peer=self.config.scan.messages.source
                    )
                    logger.info(
                        f"[DONE] Message {msg_id} forwarded successfully to {username or uid}"
                    )
                else:
                    logger.info(
                        f"[TEST] Would forward message {msg_id} to user {username or uid}"
                    )

                self.state.users.messaged.add(uid)
                logger.info(
                    f"[INFO] Total messaged users: {len(self.state.users.messaged)}"
                )
                return True

        except errors.UserIsBlockedError:
            logger.warning(f"[SKIP] User {uid} has blocked us")
        except errors.UserDeactivatedError:
            logger.warning(f"[SKIP] User {uid} account deactivated")
        except errors.FloodWaitError as e:
            logger.warning(
                f"[FLOOD] Rate limit exceeded for request {e.request.__class__.__name__} (wait {e.seconds} seconds)"
            )
            raise
        except errors.PeerFloodError as e:
            logger.error(
                f"[FLOOD] Peer flood error for request {e.request.__class__.__name__}"
            )
            raise
        except Exception as e:
            logger.error(
                f"[ERROR] Failed to process user {uid} ({e.__class__.__name__}): {str(e)}"
            )
            raise

        return False

    async def send_mode(
        self,
        client: TelegramClient,
        saved_msgs: List[int],
        shutdown_event: asyncio.Event,
    ):
        """
        Execute send mode operations.

        Args:
            client: Active Telegram client instance
            saved_msgs: List of message IDs available for forwarding
            shutdown_event: Event to signal operation shutdown

        Processes eligible users by sending configured messages with rate limiting
        and delay intervals.
        """
        sent = skipped = 0
        for username in list(
            self.state.users.eligible
        ):  # Create copy since we'll modify
            if shutdown_event.is_set():
                break

            user = await self._resolve_user(client, username)
            if not user:
                logger.warning(f"[SKIP] Could not resolve user @{username}")
                skipped += 1
                continue

            if user.id in self.state.users.messaged:
                logger.debug(f"[SKIP] User {username} already messaged")
                skipped += 1
                continue

            try:
                if await self.process_user(client, user, saved_msgs):
                    sent += 1
                    self.state.users.eligible.discard(username)
                    if self.config.scan.rapid_save:
                        await self.state.save()
                else:
                    skipped += 1

                await self.human_delay(shutdown_event)

                if (sent + skipped) % 10 == 0:
                    logger.info(f"[INFO] Progress: {sent} sent, {skipped} skipped")

            except Exception as e:
                logger.error(f"[ERROR] Failed to process @{username}: {e}")
                break  # Stop on any error to avoid flooding logs

        logger.info(f"[INFO] Total: {sent + skipped}, sent: {sent}, skipped: {skipped}")

    async def scan_mode(
        self,
        client: TelegramClient,
        saved_msgs: List[int],
        shutdown_event: asyncio.Event,
    ):
        """
        Execute scan mode operations.

        Args:
            client: Active Telegram client instance
            saved_msgs: List of message IDs available for forwarding
            shutdown_event: Event to signal operation shutdown

        Scans configured groups for eligible users, maintaining message windows
        for incremental scanning.
        """
        total_groups = len(self.config.scan.targets)
        logger.info(f"[SCAN] Starting scan of {total_groups} groups")

        for idx, group in enumerate(self.config.scan.targets, 1):
            if shutdown_event.is_set():
                break

            logger.info(f"[SCAN] Group {idx}/{total_groups}: {group}")

            # Get group ID and window
            gid = await self.get_group_id(client, group)
            if not gid:
                logger.error(f"[SCAN] Failed to resolve group {group}, skipping")
                continue
            group_info = self.state.groups.get(gid)
            highest_id = group_info.max
            lowest_id = group_info.min
            logger.info(
                f"[SCAN] Group {group} (ID: {gid}) window: highest={highest_id}, lowest={lowest_id}"
            )

            try:
                # Use min_id for newer messages (above highest_id)
                if highest_id:
                    logger.info(f"[SCAN] Scanning messages newer than {highest_id}")
                    await self._scan_messages(
                        client,
                        group,
                        saved_msgs,
                        shutdown_event,
                        min_id=highest_id,
                        ref_point="newer",
                    )

                # Use max_id for older messages (below lowest_id)
                if lowest_id:
                    logger.info(f"[SCAN] Scanning messages older than {lowest_id}")
                    await self._scan_messages(
                        client,
                        group,
                        saved_msgs,
                        shutdown_event,
                        max_id=lowest_id,
                        ref_point="older",
                    )

                # If no window exists, scan without restrictions
                if not highest_id and not lowest_id:
                    logger.info("[SCAN] No existing window, scanning all messages")
                    await self._scan_messages(client, group, saved_msgs, shutdown_event)

            except Exception as e:
                logger.error(f"[ERROR] Failed to scan group {group}: {e}")

    async def _scan_messages(
        self,
        client: TelegramClient,
        group: str,
        saved_msgs: List[int],
        shutdown_event: asyncio.Event,
        min_id: Optional[int] = None,
        max_id: Optional[int] = None,
        ref_point: str = "all",
    ) -> None:
        """
        Scan messages in a group with window tracking.

        Args:
            client: Active Telegram client instance
            group: Group identifier to scan
            saved_msgs: List of message IDs available for forwarding
            shutdown_event: Event to signal scanning shutdown
            min_id: Optional minimum message ID for window
            max_id: Optional maximum message ID for window
            ref_point: Reference point for logging ("newer", "older", or "all")

        Scans messages within specified window, processes found users, and
        updates scanning window state.
        """
        scanned = processed = skipped = errors = 0
        start_time = time.time()
        gid = await self.get_group_id(client, group)
        if not gid:
            logger.error(f"[SCAN] Failed to resolve group {group}, skipping")
            return
        group_info = self.state.groups.get(gid)

        # Use Telethon's native window support
        iter_params = {"limit": self.config.scan.limit}
        if min_id:
            iter_params["min_id"] = min_id
            iter_params["reverse"] = True
        if max_id:
            iter_params["max_id"] = max_id

        admins = await client.get_participants(group, filter=ChannelParticipantsAdmins)
        admin_ids = {admin.id for admin in admins}
        logger.debug(f"[SCAN] Found {len(admins)} admins in {group}")

        logger.debug(f"[SCAN] Message iterator params: {iter_params}")

        # Get thresholds at start
        skip_threshold = self._get_random_threshold("skip")
        scan_threshold = self._get_random_threshold("scan")
        process_threshold = self._get_random_threshold("process")

        async for msg in client.iter_messages(group, **iter_params):
            if shutdown_event.is_set():
                break

            scanned += 1
            logger.debug(
                f"[SCAN] Found message {msg.id} ({scanned}/{self.config.scan.limit})"
            )

            # Always update both bounds with each message
            group_info.max = max(msg.id, group_info.max or msg.id)
            group_info.min = min(msg.id, group_info.min or msg.id)

            # Message processing
            if not msg.sender:
                logger.debug(f"[SKIP] Message {msg.id}: No sender")
                skipped += 1
                if skipped % 50 == 0:  # Add delay every 50 skips
                    await self.human_delay(shutdown_event)
                continue

            if not isinstance(msg.sender, User):
                logger.debug(
                    f"[SKIP] Message {msg.id}: Sender type {type(msg.sender).__name__}"
                )
                skipped += 1
                if skipped % 50 == 0:
                    await self.human_delay(shutdown_event)
                continue

            user: User = msg.sender
            uid = user.id
            username = getattr(user, "username", None)

            if user.bot:
                logger.debug(
                    f"[SKIP] Message {msg.id}: User {username or uid} is a bot"
                )
                skipped += 1
                if skipped % skip_threshold == 0:
                    skip_threshold = self._get_random_threshold("skip")
                    await self.human_delay(shutdown_event)
                continue

            if uid in admin_ids:
                logger.debug(
                    f"[SKIP] Message {msg.id}: User {username or uid} is an admin"
                )
                skipped += 1
                if skipped % skip_threshold == 0:
                    skip_threshold = self._get_random_threshold("skip")
                    await self.human_delay(shutdown_event)
                continue

            if uid in self.state.users.messaged:
                logger.debug(
                    f"[SKIP] Message {msg.id}: User {username or uid} already messaged"
                )
                skipped += 1
                if skipped % skip_threshold == 0:
                    skip_threshold = self._get_random_threshold("skip")
                    await self.human_delay(shutdown_event)
                continue

            if username in self.state.users.eligible:
                logger.debug(
                    f"[SKIP] Message {msg.id}: User {username or uid} already eligible"
                )
                skipped += 1
                if skipped % skip_threshold == 0:
                    skip_threshold = self._get_random_threshold("skip")
                    await self.human_delay(shutdown_event)
                continue

            # Rate limiting with random threshold
            if scanned % scan_threshold == 0:
                scan_threshold = self._get_random_threshold("scan")
                await self.human_delay(shutdown_event)

            try:
                logger.info(
                    f"[PROCESS] Message {msg.id}: Processing user {username or uid}"
                )
                if await self.process_user(client, user, saved_msgs):
                    processed += 1
                    if self.config.scan.rapid_save:
                        await self.state.save()
                    # Only delay every N successful processes in scan mode
                    if self.config.scan.mode != ScanMode.SCAN or processed % 20 == 0:
                        await self.human_delay(shutdown_event)
                else:
                    logger.debug(
                        f"[SKIP] Message {msg.id}: User {uid} not processed (returned False)"
                    )
                    skipped += 1

                if processed % process_threshold == 0:
                    process_threshold = self._get_random_threshold("process")
                    await self.human_delay(shutdown_event)

                if processed % 10 == 0:
                    rate = processed / max(1, time.time() - start_time)
                    logger.info(f"[INFO] Processing rate: {rate*60:.1f} users/minute")
                    await self.state.save()  # Regular state saves

            except Exception as e:
                logger.error(
                    f"[ERROR] Message {msg.id}: Failed to process user {uid}: {str(e)}"
                )
                errors += 1
                break

        elapsed = time.time() - start_time
        if errors > 0:
            logger.error(
                f"[ERROR] {ref_point.title()} messages scan encountered {errors} errors"
            )
        else:
            logger.info(
                f"[COMPLETE] {ref_point.title()} messages scan finished in {elapsed:.1f}s:"
            )
        logger.info(
            f"[INFO] Scanned: {scanned}, Processed: {processed}, Skipped: {skipped}"
        )
        if processed > 0:
            logger.info(f"[INFO] Rate: {(processed/elapsed)*60:.1f} users/minute")

        # Save state after each direction scan
        await self.state.save()

    async def receive_mode(
        self, client: TelegramClient, shutdown_event: asyncio.Event
    ) -> None:
        """
        Execute receive mode operations.

        Args:
            client: Active Telegram client instance
            shutdown_event: Event to signal operation shutdown

        Receives and displays new messages from the configured source chat.
        """
        logger.info("[RECEIVE] Starting to receive new messages")

        @client.on(events.NewMessage)
        async def handler(event: events.NewMessage.Event):
            msg: Message = event.message
            user = await client.get_entity(msg.peer_id)
            logger.info(f"[MESSAGE] From {user.id}\n{msg.message}")

            # mark message as read
            await client(
                functions.messages.ReadHistoryRequest(peer=msg.peer_id, max_id=msg.id)
            )

        # Wait for shutdown event
        await shutdown_event.wait()

        # Remove event handler
        client.remove_event_handler(handler)
        logger.info("[RECEIVE] Stopped receiving messages")

    async def _resolve_exclusions(
        self, client: TelegramClient, shutdown_event: asyncio.Event
    ) -> None:
        """Resolve excluded usernames to UIDs and add to messaged state."""
        total = resolved = 0
        for username in self.config.exclusions:
            total += 1
            user = await self._resolve_user(client, username)

            if shutdown_event.is_set():
                return

            if user:
                if username in self.state.users.eligible:
                    self.state.users.eligible.discard(username)
                    logger.info(f"[USER] Removed @{username} from eligible list")
                if user.id not in self.state.users.messaged:
                    self.state.users.messaged.add(user.id)
                    logger.info(f"[USER] Added {user.id} to messaged list")
                else:
                    logger.debug(f"[SKIP] {user.id} already in messaged list")
                resolved += 1
            else:
                logger.warning(f"[USER] Failed to resolve @{username}")

            if total % 15 == 0:
                await self.human_delay(shutdown_event)
        logger.info(f"[INFO] Resolved {resolved}/{total} excluded users")

    async def run(self, shutdown_event: asyncio.Event):
        """
        Execute main scanner operations.

        Args:
            shutdown_event: Event to signal operation shutdown

        Sets up proxy connection, initializes client, and executes operations
        based on configured mode.
        """
        try:
            # Get session
            session = utils.get_session(self.config)

            # Test proxies and get working one
            self.app.status.update("Testing connections...")
            logger.info("Testing connections")
            await self.proxy_manager.test_proxies(shutdown_event=shutdown_event)
            working_connection = self.proxy_manager.current_proxy

            if shutdown_event.is_set():
                return

            if working_connection:
                if working_connection.is_direct:
                    logger.info("Using direct connection")
                else:
                    self.app.console.print(
                        f"Using {working_connection.type.value} proxy: {working_connection.host}:{working_connection.port}"
                    )
                    logger.info(
                        f"Using proxy: [{working_connection.type.name}] {working_connection.host}:{working_connection.port}"
                    )
            else:
                self.app.console.print("No working connection available, aborting")
                logger.error("No working connection available, aborting")
                return

            self.app.status.update("Initializing client...")
            # Create actual client with working connection
            client: TelegramClient = utils.get_client(
                config=self.config, proxy=working_connection, session=session
            )

            self.app.status.update("Logging in...")
            logger.info(f"Logging in with {client.session.filename} session")

            # Continue with client operations
            async with client:
                if shutdown_event.is_set():
                    return

                self.app.status.update("Fetching account information...")
                logger.debug("Fetching account information")
                me: User = await client.get_me()

                if me.first_name and me.last_name:
                    name = f"{me.first_name} {me.last_name}"
                else:
                    name = me.first_name

                self.app.status.update("Updating status...")
                logger.debug("Updating status to online")
                result = await client(
                    functions.account.UpdateStatusRequest(offline=False)
                )
                if not result:
                    self.app.console.print(
                        "Failed to set status to online", style=Style(color="red")
                    )
                    logger.error("Failed to set status to online")
                else:
                    logger.debug("Status set to online successfully")

                self.app.console.print(f"Logged in as {name} (ID: {me.id})")
                logger.info(f"Logged in as {name} (ID: {me.id})")

                # Log detailed account info for debugging
                if self.config.app.verbose:
                    logger.debug(f"Account ID: {me.id}")
                    logger.debug(f"Full Name: {name}")
                    logger.debug(f"Username: {me.username}")
                    logger.debug(f"Phone: {me.phone}")

                # Show unread messages count
                self.app.status.update("Checking unread messages...")
                logger.debug("Checking for unread messages")
                total_unread = 0
                async for dialog in client.iter_dialogs(limit=50):
                    logger.debug(f"Processing {dialog.name} ({dialog.id})")
                    try:
                        unread = int(getattr(dialog, "unread_count", 0) or 0)
                    except Exception:
                        unread = 0
                    logger.debug(f"Unread count: {unread}")
                    if unread <= 0:
                        continue
                    muted = self._is_dialog_muted(dialog)
                    logger.debug(f"Is muted: {muted}")
                    if muted:
                        continue
                    total_unread += unread

                if total_unread > 0:
                    self.app.console.print(
                        f"You have {total_unread} unread messages",
                        style=Style(color="yellow"),
                    )
                    logger.warning(f"You have {total_unread} unread messages")
                else:
                    logger.debug("No unread messages")

                try:
                    await self.start(client, shutdown_event)
                finally:
                    self.app.status.update("Shutting down client...")
                    if not client.is_connected():
                        if self.config.scan.mode != ScanMode.LOGOUT:
                            logger.warning("Client disconnected unexpectedly")

                        return

                    logger.debug("Updating status to offline")
                    result = await client(
                        functions.account.UpdateStatusRequest(offline=True)
                    )
                    if not result:
                        logger.error("Failed to set status to offline")
                    else:
                        logger.debug("Status set to offline successfully")

            logger.info("Client session closed")

        except SessionNotFoundError as e:
            self.app.console.print(
                "Session not found:",
                self.config.telegram.session,
                style=Style(color="red"),
            )
            logger.error(f"Session not found: {e}")
        except Exception as e:
            logger.error(f"Unexpected error: {e}")
            raise

    async def start(self, client: TelegramClient, shutdown_event: asyncio.Event):
        self.app.status.update("Starting operations...")

        if self.config.scan.mode == ScanMode.HALT:
            return

        # Account logout
        if self.config.scan.mode == ScanMode.LOGOUT:
            user_input = self.app.console.input(
                "Are you sure you want to log out and delete the session? (y/N): "
            )
            if user_input.lower() == "y":
                self.app.status.update("Logging out...")
                logger.info("Logging out and deleting session")
                result = await client.log_out()
                if result:
                    self.app.console.print(
                        "Logged out successfully", style=Style(color="green")
                    )
                    logger.info("Logged out successfully")
                else:
                    self.app.console.print(
                        "Failed to log out", style=Style(color="red")
                    )
                    logger.error("Failed to log out")
            else:
                self.app.console.print("Logout canceled", style=Style(color="yellow"))
                logger.info("Logout canceled")
            return

        # Process exclusions if any
        if hasattr(self.config, "exclusions") and self.config.exclusions:
            await self._resolve_exclusions(client, shutdown_event)
            return

        if self.config.scan.mode == ScanMode.RECEIVE:
            await self.receive_mode(client, shutdown_event)
            return

        saved_msgs = await self.get_saved_messages(client)
        if not saved_msgs:
            logger.error("No saved messages available")
            return

        if self.config.scan.mode == ScanMode.SEND:
            await self.send_mode(client, saved_msgs, shutdown_event)
        else:
            await self.scan_mode(client, saved_msgs, shutdown_event)

    @staticmethod
    def _is_dialog_muted(dialog: Dialog) -> bool:
        try:
            if not dialog.dialog.notify_settings.mute_until:
                return False

            return (
                dialog.dialog.notify_settings.mute_until.timestamp()
                > datetime.now().timestamp()
            )
        except Exception:
            return False

    delay_variation: int
    """Maximum variation in seconds to add/subtract from base thresholds"""

    _thresholds: Dict[str, int]
    """Base thresholds for different operation types:
    - skip: Messages to skip before delay
    - scan: Messages to scan before delay
    - process: Users to process before delay
    """

    _cached_ids: Dict[str, str]
    """Cache of resolved group identifiers to numeric IDs"""

    _entity_cache: Dict[str, User]
    """Cache of resolved usernames to User entities"""



from . import utils
from .__version__ import __version__
from .app import Application
from .config import Config, InteractiveMode, ScanMode
from .exceptions import SessionMismatchError, SessionNotFoundError
from .logger import setup_logger
from .proxy_manager import Proxy, ProxyManager
from .scanner import TelegramScanner
from .session import CustomSession
from .state_manager import StateManager

__all__ = [
    "__version__",
    "Application",
    "Config",
    "ScanMode",
    "TelegramScanner",
    "StateManager",
    "setup_logger",
    "CustomSession",
    "SessionNotFoundError",
    "SessionMismatchError",
    "utils",
    "ProxyManager",
    "Proxy",
    "InteractiveMode",
]



import typer

from ..app import Application
from ..config import Config
from ..session import CustomSession

app = typer.Typer(
    help="Manage Telegram sessions",
)


@app.command("login")
def login(
    ctx: typer.Context,
    session: str = typer.Argument(
        None,
        help="Session name for authentication",
    ),
):
    """Login to a Telegram session."""
    from .. import tasks

    config: Config = ctx.obj["g_load_config"]()
    if session:
        config.telegram.session = session
    app: Application = ctx.obj["g_load_app"]()
    app.start(tasks.session.login)


@app.command("list")
def list_sessions(
    ctx: typer.Context,
):
    """List all available Telegram sessions."""
    ctx.obj["g_load_config"]()
    sessions = CustomSession.list_sessions()
    if sessions:
        typer.echo("Available sessions:")
        for s in sessions:
            typer.echo(s)
    else:
        typer.echo("No sessions found.")
    raise typer.Exit()



import time
from pathlib import Path
from typing import List, Optional

import typer
from click import FileError

from . import commands, utils
from .app import Application
from .config import Config, ScanMode, find_config_file
from .logger import setup_logger


def validate_config_path(value: Path) -> Path:
    """
    Validate configuration file path.

    Args:
        value: Path to validate

    Returns:
        Path: Valid configuration file path

    Raises:
        typer.BadParameter: If path doesn't exist or points to directory
    """
    resolved_path = value.resolve()
    if not resolved_path.exists():
        raise typer.BadParameter(f"Config file does not exist: {resolved_path}")
    if resolved_path.is_dir():
        raise typer.BadParameter(f"Config path points to a directory: {resolved_path}")

    return resolved_path


app = typer.Typer(
    help="CLI based telegram client",
    add_completion=False,
)

app.add_typer(commands.config.app, name="config")
app.add_typer(commands.session.app, name="session")
app.add_typer(commands.proxy.app, name="proxy")


@app.command()
def run(
    ctx: typer.Context,
    test: bool = typer.Option(
        False, "-t", "--test", help="Run in test mode without sending actual messages"
    ),
    rapid_save: bool = typer.Option(
        False, "--rapid-save", help="Enable rapid state saving to disk"
    ),
    mode: ScanMode = typer.Option(
        ScanMode.FULL.value, "-m", "--mode", help="Operation mode"
    ),
    session: Optional[str] = typer.Option(
        None, "-s", "--session", help="Telethon session name for authentication"
    ),
    limit: Optional[int] = typer.Option(
        None, "-l", "--limit", help="Maximum number of messages to process per group"
    ),
    exclude: Optional[Path] = typer.Option(
        None,
        "-e",
        "--exclude",
        help="JSON file with usernames to exclude from processing",
    ),
):
    """Telegram message scanner and forwarder."""
    typer.echo("The 'run' command is currently under development.")
    typer.Exit(1)

    # config: Config = ctx.obj["g_load_config"]()
    # if test:
    #     config.scan.test = True
    # if rapid_save:
    #     config.scan.rapid_save = True
    # if mode:
    #     config.scan.mode = mode
    # if session:
    #     config.telegram.session = session
    # if limit is not None:
    #     config.scan.limit = limit
    # if exclude:
    #     config.exclusions = json.load(exclude.open("r"))
    # app = Application(config=config)
    # app.start()


@app.command("interactive")
def interactive(
    ctx: typer.Context,
    session: Optional[str] = typer.Option(
        None,
        "-s",
        "--session",
        help="Session name for authentication",
    ),
):
    """Run the application in interactive mode."""
    from .tasks import interactive

    config: Config = ctx.obj["g_load_config"]()
    if session:
        config.telegram.session = session

    app: Application = ctx.obj["g_load_app"]()
    app.start(interactive.main)


@app.command("info")
def info():
    """Display information about cligram and current environment."""
    from . import __version__

    typer.echo(f"cligram version: {__version__}")

    device_info = utils.get_device_info()
    typer.echo(f"Operating System: {device_info.os.value}")
    typer.echo(f"Architecture: {device_info.architecture.value}")
    typer.echo(f"OS Title: {device_info.os_title}")
    typer.echo(f"OS Version: {device_info.version}")
    typer.echo(f"Device Model: {device_info.model}")


@app.callback()
def callback(
    ctx: typer.Context,
    config: Optional[Path] = typer.Option(
        None,
        "-c",
        "--config",
        help="Path to JSON configuration file",
    ),
    verbose: bool = typer.Option(
        False, "-v", "--verbose", help="Enable detailed debug logging output to console"
    ),
    overrides: List[str] = typer.Option(
        [],
        "-o",
        "--override",
        help="Override config values using dot notation (e.g., app.verbose=true)",
    ),
):
    """
    CLI entry point for cligram application.
    """
    ctx.obj = {}

    def setup() -> Config:
        nonlocal config, verbose, overrides

        if Config.get_config(raise_if_failed=False) is not None:
            return Config.get_config()

        config = config or find_config_file(raise_error=True)
        if not config:
            raise FileError("Configuration file not found.")

        loaded_config = Config.from_file(config, overrides=overrides)
        if verbose:
            loaded_config.app.verbose = True

        setup_logger(
            verbose=loaded_config.app.verbose,
            log_file=loaded_config.base_path
            / "logs"
            / time.strftime("%Y-%m-%d.log", time.localtime()),
        )

        return loaded_config

    ctx.obj["g_load_config"] = setup

    def load_app() -> Application:
        from .app import Application

        cfg = ctx.obj["g_load_config"]()
        return Application(config=cfg)

    ctx.obj["g_load_app"] = load_app


def main():
    app()



class SessionMismatchError(Exception):
    """Raised when the session's API ID does not match the configured API ID."""

    pass


class SessionNotFoundError(Exception):
    """Raised when the specified session file is not found."""

    pass


class NoWorkingConnectionError(Exception):
    """Raised when no working connection (direct or proxy) is available."""

    pass
